# TinyTuya Contrib ThermostatDevice Module
# -*- coding: utf-8 -*-
"""
 A community-contributed Python module to add support for Tuya WiFi smart thermostats

 This module attempts to provide everything needed so there is no need to import the base tinytuya module

 Module Author: uzlonewolf (https://github.com/uzlonewolf)

 Local Control Classes
    ThermostatDevice(..., version=3.3, persist=True)
        This class uses a default version of 3.3 and enables persistance so we can catch temperature updates
        See OutletDevice() for the other constructor arguments

 Additional Classes
    ThermostatSensorList(dps, parent_device)
        Mainly used internally, exposed in case it's useful elsewhere
        The 'dps' argument should be the DPS ID of the list so it knows what DPS to send when updating a sensor option
        The 'parent_device' argument should be the ThermostatDevice() this sensor list belongs to


    Sensor related functions:
        tstatdev = ThermostatDevice(...)
        tstatdev.sensors
            -> an iterable list of all the sensors that can also be acessed like a dict:

              for sensor in tstatdev.sensors:
                  if not sensor.online:
                      print( 'Sensor %r offline!' % sensor.name )

              if tstatdev.sensors['12345678'].battery < 10:
                  print( 'Sensor %r low battery!' % tstatdev.sensors['12345678'].name )
              else:
                  print( 'Sensor %r battery %d%%' % (tstatdev.sensors['12345678'].name, tstatdev.sensors['12345678'].battery ) )

            dict access matches against both name and id:
              tstatdev.sensors['name'] and tstatdev.sensors['id'] both work

        When sensor values change, the sensor object is also available in data['changed_sensors'].  i.e.
            data = tstatdev.receive()
            if data and 'changed_sensors' in data:
                for sensor in data['changed_sensors']:
                    if 'temperature' in sensor['changed'] and sensor.online:
                        ...do something with sensor.temperature or whatever...

        sensors have the following attributes:
          id -> ID # of the sensor as a hex string
          raw_id -> ID # of the sensor as a integer
          name -> decoded and trimmed name of the sensor
          raw_name -> NUL-padded name of the sensor as a byte array
          enabled -> sensor enabled flag True/False
          occupied -> sensor detected occupancy flag True/False
          temperature -> temperature in degrees C as a float
          raw_temperature -> temperature as reported by the sensor (degrees C * 100)
          temperature_used -> the rounded temperature used in averaging calculations
          raw_temperature_used -> the rounded temperature used in averaging calculations (degrees C * 100)
          online -> sensor online flag True/False
          participation -> schedule participation bitmask ['wake', 'away', 'home', 'sleep']
          battery -> battery percentage remaining
          unknown2 -> value of unknown field, integer 0-255
          firmware_version -> firmware version * 10 (01 = v0.1)
          averaging -> sensor currently participation in the temperature averaging (occupied is True and participation flag for the current schedule mode is set)
          unknown3 -> value of unknown field, 8 byte long byte array
          changed -> list of attributes which have changed since last update

        sensors also have the following methods:
          sensor.setName( new_name )
          sensor.setEnabled( enabled )
          sensor.setOccupied( occupied )
              -> not really useful for remote sensors as they get overwritten on the next update
          sensor.setParticipation( flag, val=True )
              -> flag can be either a string in ['wake', 'away', 'home', 'sleep'] or an integer bitmask
                 when it's a string, val sets (True) or clears (False) that particular flag
                 when it's a integer, the bitmask is set to flag and val is ignored
          sensor.getParticipation( flag )
              -> flag can be either a string in ['wake', 'away', 'home', 'sleep'] or an integer bitmask
                 returns True if (string) flag is set or (integer) bitmask matches exactly, otherwise returns False
              -> if the current value of all flags is wanted, the sensor.participation field can be read directly instead of using this function
          sensor.setUnknown2( val )
              -> sets the second unknown field to val.  'val' should be an integer in the range 0-255
          sensor.setUnknown3( val )
              -> sets the third unknown field to val.  'val' should be a 8 byte long byte array

        If multiple sensor options are going to be changed at the same time, it is much quicker to queue the updates and send them all at once:
            sensor.delayUpdates()
            ... call sensor.setName() or whatever here ...
            sensor.sendUpdates()


    Thermostat related functions:
        delayUpdates()
            -> when changing multiple settings, calling this first will cause them to be queued and sent all at once later
        sendUpdates()
            -> sends all queued updates at once and disables queueing (delayUpdates() will need to be called again if you want to queue things)

        setSetpoint( setpoint, cf=None )
            -> tried to auto-detect which setpoint you want to set (cooling or heating)) using the system mode and sets it
               if cf is None it assumes the given setpoint is the same temperature unit (degrees C or F) as the system temperature unit
        setCoolSetpoint( setpoint, cf=None )
            -> sets the cooling setpoint, for when the system mode is 'cool' or 'auto'
        setHeatSetpoint( setpoint, cf=None )
            -> sets the heating setpoint, for when the system mode is 'heat' or 'auto'
        setMiddleSetpoint( setpoint, cf=None )
            -> you should not need to call this, the thermostat handles it
               matches the cool or heat setpoint if the system is in those modes, or the midpoint between them if the mode is 'auto'

        setMode( mode )
            -> sets the system mode.  mode should be a string in ['cool', 'heat', 'auto', 'off']
        setFan( fan )
            -> sets the fan mode.  fan should be True (on), False (auto), or a string in ['on', 'auto', 'circ']
        setFanRuntime( runtime )
            -> when the fan mode is 'circ' this sets how many minutes per hour the fan is run to circulate the air
        setUnits( cf )
            -> sets the system temperature units.  cf should be a string in ['c', 'f']
        setSchedule( schedule )
            -> enables or disables the previously-created schedule.  creating a new schedule is not yet implemented
        setHold( hold )
            -> sets the temperature hold.  hold should be True (permhold), False (followschedule), or a string in ['permhold', 'temphold', 'followschedule']

        getCF(cf=None)
            -> parses the given cf value and returns either 'c' or 'f', or returns the system temperature units if cf is None
        isSingleSetpoint()
            -> returns True if the system is expecting a single temperature (mode is 'cool' or 'heat'), or False if it is expecting separate cool and heat setpoints

        sendPing()
            -> sends a async heartbeat packet
        sendStatusRequest()
            -> sends a async status request packet
        status()
            -> sends a synchronous status request packet and returns the result after parsing it
        receive()
            -> receives a single packet and returns the result after parsing it

        setValue( key, val )
            -> directly set a key in the dict.  you probably do not need to call this directly
        setValues( dict )
            -> directly set multiple keys in the dict.  you probably do not need to call this directly
        parseValue( key, val )
            -> converts a value to the format the DPS is expecting for that particular key.  you probably do not need to call this directly

    attributes:
        mode -> ['auto', 'cool', 'heat', 'emergencyheat', 'off']
        fan -> ['auto', 'cycle', 'on']
        system -> current system state, ['fanon', 'coolfanon', 'alloff', 'heatfanon', 'heaton']
        setpoint_c -> either the setpoint when system is not in 'auto' mode, or the midpoint between the heating and cooling setpoints
        temp_set -> alias for setpoint_c
        setpoint_f and temp_set_f -> same as setpoint_c but in degrees F
        cooling_setpoint_c and upper_temp
        cooling_setpoint_f and upper_temp_f
        heating_setpoint_c and lower_temp
        heating_setpoint_f and lower_temp_f
        units and temp_unit_convert -> system temperature units, either 'c' or 'f'
        temp_correction -> offset to adjust displayed sensor temperatures
        temperature_c and temp_current -> current temperature in degrees C
        temperature_f and temp_current_f -> current temperature in degrees F
        humidity -> RH%
        fault -> fault flags, [e1, e2, e3]
        system_type -> '4'=heatpump, '5'=2-stage heatpump?
        home -> ??
        schedule -> ThermostatSchedule() class
        schedule_enabled -> flag True/False
        hold -> ['permhold', 'temphold', 'followschedule']
        vacation -> binary blob
        fan_run_time -> when the fan mode is 'circ' this is how many minutes per hour the fan is run to circulate the air
        weather_forcast -> ??

    status() and receive() both return a dict containing both the raw DPS dict as well as a list of changed attributes in 'changed' and a list of changed sensors in 'changed_sensors'
     these can be used like:

            data = tstatdev.receive()
            if data and 'changed' in data:
                if 'system' in data['changed']:
                    print( 'System State changed, current temperature is:', tstatdev.temperature_c )
                for changed in data['changed']:
                    print( 'Changed:', changed, 'New Value:', getattr( tstatdev, changed ) )
            if data and 'changed_sensors' in data:
                for sensor in data['changed_sensors']:
                    print( 'Sensor Changed! Changed Attribs:%r DPS:%s ID:%s Name:"%s" Current Temperature: %r' % (sensor.changed, sensor.parent_sensorlist.dps, sensor.id, sensor.name, sensor.temperature) )
                    if 'sensor_added' in sensor.changed:
                        print( 'New sensor was added!' )
                    if 'name' in sensor.changed:
                        print( 'Sensor was renamed!' )
                    for changed in sensor.changed:
                        print( 'Changed:', changed, 'New Value:', getattr( sensor, changed ) )

    ThermostatSchedule class:

        !! WARNING !! The thermostat does NOT send the current schedule when you request the status, it only sends it when it has changed.  So, you
            must either a) set the entire schedule or b) change it in the app or on the thermostat itself while tinytuya is running.  Changes to a
            single day/period/value can only be made once this has been done.

        'day' is a case-insensitive string starting with su, m, tu, w, th, f, sa or an integer in the range 0-6
        'period' is a case-insensitive string starting with w[ake], a[way], h[ome], s[leep], e[xtra] or an integer in the range 0-4
            -> Only periods 0-3 (wake-sleep) show up in the app or on the thermostat! (4 (extra) is hidden)

        Schedule parameters can be accessed directly by dict via name or index:
            tstatdev.schedule[1][0].coolto = 25.0 or
            tstatdev.schedule['monday']['wake'].coolto = 25.0 or
            tstatdev.schedule['m']['w'].coolto = 25.0 or
            tstatdev.schedule['MoNdAySsUcK']['WakeMeUp'].coolto = 25.0
        all mean the same thing.  Parameters can also be set using the .setPeriod method:
            tstatdev.schedule.setPeriod( day_of_week, period, coolto=25.0, heatto=10.0, time=0, participation=(period & 3) )
        To disable a schedule period (set the time to 0xFFFF) you can:
            tstatdev.schedule.setPeriod( day, 4, delete=True)

        Once a day is set you can copy it to a different day with:
            # copy sunday (0) to monday-saturday (1-6)
            for i in range(6):
                tstatdev.schedule.copyDay( 0, i+1 )

        Individual periods can also be copied:
            tstatdev.schedule.copyPeriod( src_day, src_period, dst_day, dst_period )
"""

import struct
import base64

from ..core import Device, log, HEART_BEAT, DP_QUERY, CONTROL

class ThermostatDevice(Device):
    """
    Represents a Tuya based 24v Thermostat.
    """

    sensor_dps = ('122', '125', '126', '127', '128')
    dps_data = {
        '2' : { 'name': 'mode', 'enum': ['auto', 'cool', 'heat', 'emergencyheat', 'off'] },
        '16': { 'name': 'temp_set', 'alt': 'setpoint_c', 'scale': 100 },
        '17': { 'name': 'temp_set_f', 'alt': 'setpoint_f' },
        '18': { 'name': 'upper_temp_f', 'alt': 'cooling_setpoint_f', 'high_resolution': False },
        '19': { 'name': 'upper_temp', 'alt': 'cooling_setpoint_c', 'high_resolution': False },
        '20': { 'name': 'lower_temp_f', 'alt': 'heating_setpoint_f', 'high_resolution': False },
        '23': { 'name': 'temp_unit_convert', 'alt': 'units', 'enum': ['f','c'] },
        '24': { 'name': 'temp_current', 'alt': 'temperature_c', 'scale': 100 },
        '26': { 'name': 'lower_temp', 'alt': 'heating_setpoint_c', 'high_resolution': False },
        '27': { 'name': 'temp_correction', 'high_resolution': False },
        '29': { 'name': 'temp_current_f', 'alt': 'temperature_f' },
        '34': { 'name': 'humidity' },
        '45': { 'name': 'fault' },
        '107': { 'name': 'system_type', 'decode': int },
        '108': { 'name': 'upper_temp', 'alt': 'cooling_setpoint_c', 'scale': 100, 'high_resolution': True },
        '109': { 'name': 'lower_temp', 'alt': 'heating_setpoint_c', 'scale': 100, 'high_resolution': True },
        '110': { 'name': 'upper_temp_f', 'alt': 'cooling_setpoint_f', 'high_resolution': True },
        '111': { 'name': 'lower_temp_f', 'alt': 'heating_setpoint_f', 'high_resolution': True },
        '115': { 'name': 'fan', 'enum': ['auto', 'cycle', 'on'] },
        '116': { 'name': 'home' },
        '118': { 'name': 'schedule', 'base64': True, 'selfclass': 'ThermostatSchedule' },
        '119': { 'name': 'schedule_enabled' },
        '120': { 'name': 'hold', 'enum': ['permhold', 'temphold', 'followschedule'] },
        '121': { 'name': 'vacation', 'base64': True },
        #'122': { 'name': 'sensor_list_1', 'base64': True },
        '123': { 'name': 'fan_run_time' }, # presumably for when fan='circ'
        #'125': { 'name': 'sensor_list_2', 'base64': True },
        #'126': { 'name': 'sensor_list_3', 'base64': True },
        #'127': { 'name': 'sensor_list_4', 'base64': True },
        #'128': { 'name': 'sensor_list_5', 'base64': True },
        '129': { 'name': 'system', 'enum': ['fanon', 'coolfanon', 'alloff', 'heatfanon', 'heaton'] },
        '130': { 'name': 'weather_forcast' }
        }

    def __init__(self, *args, **kwargs):
        # set the default version to 3.3 as there are no 3.1 devices
        if 'version' not in kwargs or not kwargs['version']:
            kwargs['version'] = 3.3
        # set persistant so we can receive sensor broadcasts
        if 'persist' not in kwargs:
            kwargs['persist'] = True
        super(ThermostatDevice, self).__init__(*args, **kwargs)

        self.high_resolution = None
        self.schedule = None
        self.delay_updates = False
        self.delayed_updates = { }
        self.sensorlists = [ ]
        self.sensors = self.SensorList( self )

        for k in self.sensor_dps:
            self.sensorlists.append(ThermostatSensorList(k, self))

        for k in self.dps_data:
            val = None

            if 'selfclass' in self.dps_data[k]:
                val = getattr( self, self.dps_data[k]['selfclass'] )( self, k )

            setattr( self, self.dps_data[k]['name'], val )
            if 'alt' in self.dps_data[k]:
                setattr( self, self.dps_data[k]['alt'], val )

            if( ('scale' in self.dps_data[k]) or (('base64' in self.dps_data[k]) and self.dps_data[k]['base64']) or ('selfclass' in self.dps_data[k]) or ('decode' in self.dps_data[k]) ):
                self.dps_data[k]['check_raw'] = True

            if 'check_raw' in self.dps_data[k] and self.dps_data[k]['check_raw']:
                setattr( self, 'raw_' + self.dps_data[k]['name'], None )

    def delayUpdates( self ):
        self.delay_updates = True

    def setSetpoint( self, setpoint, cf=None ):
        if self.mode == 'cool':
            return self.setCoolSetpoint( setpoint, cf )
        elif self.mode == 'heat' or self.mode == 'emergencyheat':
            return self.setHeatSetpoint( setpoint, cf )
        else:
            # no idea, let the thermostat figure it out
            return self.setMiddleSetpoint( setpoint, cf )

    def setCoolSetpoint( self, setpoint, cf=None ):
        k = 'cooling_setpoint_' + self.getCF( cf )
        return self.setValue( k, setpoint )

    def setHeatSetpoint( self, setpoint, cf=None ):
        k = 'heating_setpoint_' + self.getCF( cf )
        return self.setValue( k, setpoint )

    def setMiddleSetpoint( self, setpoint, cf=None ):
        k = 'setpoint_' + self.getCF( cf )
        return self.setValue( k, setpoint )

    def setMode( self, mode ):
        return self.setValue( 'mode', mode )

    def setFan( self, fan ):
        if not fan:
            fan = 'auto'
        elif fan is True:
            fan = 'on'
        return self.setValue( 'fan', fan )

    def setUnits( self, cf ):
        cf = self.getCF( cf )
        return self.setValue( 'temp_unit_convert', cf )

    def setSchedule( self, sch ):
        # FIXME set schedule data?
        if sch:
            return self.setValue( 'schedule_enabled', True )
        return self.setValue( 'schedule_enabled', False )

    def setHold( self, hold ):
        if hold is True:
            return self.setValue( 'hold', 'permhold' )

        if hold is False:
            return self.setValue( 'hold', 'followschedule' )

        return self.setValue( 'hold', hold )

    def setFanRuntime( self, rt ):
        return self.setValue( 'fan_run_time', int(rt) )

    def setValue( self, key, val ):
        dps, val = self.parseValue( key, val )

        if not self.delay_updates:
            return self.set_value( dps, val, nowait=True )

        self.delayed_updates[dps] = val
        return True

    def setValues( self, val_dict ):
        for key in val_dict:
            dps, val = self.parseValue( key, val_dict[key] )
            self.delayed_updates[dps] = val

        if not self.delay_updates:
            payload = self.generate_payload(CONTROL, self.delayed_updates)
            self.delayed_updates = { }
            return self.send(payload)

        return True

    def parseValue( self, key, val ):
        dps = None
        for k in self.dps_data:
            if( (key == self.dps_data[k]['name']) or (('alt' in self.dps_data[k]) and (key == self.dps_data[k]['alt'])) ):
                if( ('high_resolution' not in self.dps_data[k]) or (self.dps_data[k]['high_resolution'] == self.high_resolution) ):
                    dps = k
                    break

        if not dps:
            log.warn( 'Requested key %r not found!' % key )
            return False

        ddata = self.dps_data[dps]

        if 'scale' in ddata:
            val = int( val * ddata['scale'] )

        if 'encode' in ddata:
            val = ddata['encode']( val )

        if 'enum' in ddata:
            if val not in ddata['enum']:
                log.warn( 'Requested value %r for key %r/%r not in enum list %r !  Setting anyway...' % (val, dps, key, ddata['enum']) )

        if 'base64' in ddata:
            val = base64.b64encode( val ).decode('ascii')

        return ( dps, val )

    def sendUpdates( self ):
        self.delay_updates = False

        if len(self.delayed_updates) > 0:
            payload = self.generate_payload(CONTROL, self.delayed_updates)
            self.delayed_updates = { }
            return self.send(payload)

        return False

    def getCF( self, cf=None ):
        if cf is None:
            cf = getattr(self, 'temp_unit_convert', 'c')
        if cf == 'f':
            return 'f'
        return 'c'

    def isSingleSetpoint( self ):
        if self.mode == 'auto':
            return False

        return True

    def sendPing( self ):
        payload = self.generate_payload( HEART_BEAT )
        return self.send(payload)

    def sendStatusRequest( self ):
        payload = self.generate_payload( DP_QUERY )
        return self.send(payload)

    def status(self):
        return super(ThermostatDevice, self).status()

    def receive(self):
        return self._send_receive(None)

    def _process_response( self, data ):
        if not data:
            return data

        if 'dps' not in data:
            return data

        data['changed'] = [ ]
        data['changed_sensors'] = [ ]

        for i in range( len(self.sensor_dps) ):
            k = self.sensor_dps[i]
            if k in data['dps']:
                data['changed_sensors'] += self.sensorlists[i].update( data['dps'][k] )

        if self.high_resolution is None:
            for k in self.dps_data:
                if k in data['dps'] and 'high_resolution' in self.dps_data[k]:
                    self.high_resolution = self.dps_data[k]['high_resolution']
                    log.info('ThermostatDevice: high-resolution is now %r' % self.high_resolution)
                    break

        for k in data['dps']:
            if k in self.dps_data:
                name = self.dps_data[k]['name']
                checkname = ('raw_' + name) if 'check_raw' in self.dps_data[k] and self.dps_data[k]['check_raw'] else name
                val = data['dps'][k]

                if getattr( self, checkname ) == val:
                    continue

                data['changed'].append( name )
                if name != checkname: data['changed'].append( checkname )
                setattr( self, checkname, val )

                if ('base64' in self.dps_data[k]) and self.dps_data[k]:
                    val = base64.b64decode( val )

                if 'selfclass' in self.dps_data[k]:
                    getattr( self, name ).update( val )

                    if 'alt' in self.dps_data[k]:
                        data['changed'].append( self.dps_data[k]['alt'] )
                        setattr( self, self.dps_data[k]['alt'], getattr( self, name ) )
                else:
                    if 'decode' in self.dps_data[k]:
                        val = self.dps_data[k]['decode']( val )

                    if 'scale' in self.dps_data[k]:
                        val /= self.dps_data[k]['scale']

                    setattr(self, name, val)

                    if 'enum' in self.dps_data[k]:
                        if val not in self.dps_data[k]['enum']:
                            log.warn( 'Received value %r for key %r/%r not in enum list %r !  Perhaps enum list needs to be updated?' % (val, k, name, self.dps_data[k]['enum']) )

                    if 'alt' in self.dps_data[k]:
                        data['changed'].append( self.dps_data[k]['alt'] )
                        setattr( self, self.dps_data[k]['alt'], val )

        return data

    def __iter__(self):
        for k in self.dps_data:
            if 'alt' in self.dps_data[k]:
                yield (self.dps_data[k]['alt'], getattr(self, self.dps_data[k]['alt']))
            yield (self.dps_data[k]['name'], getattr(self, self.dps_data[k]['name']))

    class SensorList:
        def __init__( self, parent ):
            self.parent = parent

        def find_sensor( self, name ):
            for l in self.parent.sensorlists:
                for s in l:
                    if s.id == name or s.name == name:
                        return s

            return None

        def __getitem__( self, key ):
            if isinstance( key, str ):
                return self.find_sensor( key )
            elif not isinstance( key, int ):
                return getattr( self, key )

            i = 0
            for l in self.parent.sensorlists:
                for s in l:
                    if i == key:
                        return s
                    i += 1

            return None

        def __len__( self ):
            i = 0
            for l in self.parent.sensorlists:
                for s in l:
                    i += 1
            return i

        def __iter__( self ):
            for l in self.parent.sensorlists:
                for s in l:
                    yield s

        def __call__( self ):
            for l in self.parent.sensorlists:
                for s in l:
                    yield s


    class ThermostatSchedule(object):
        class ScheduleDay:
            class SchedulePeriod:
                def __init__( self, sched ):
                    self.sched = sched
                    self.participation = 0xFF
                    self.time = 0xFFFF
                    self.heatto = -32768
                    self.coolto = -32768

                def __setitem__( self, key, data ):
                    if not isinstance( key, int ):
                        setattr( self, key, data )

                    if key == 0: self.participation = data
                    elif key == 1: self.time = data
                    elif key == 2: self.heatto = data
                    elif key == 3: self.coolto = data
                    else: raise IndexError('Numeric index must be an integer 0-3')

                def __getitem__( self, key ):
                    if not isinstance( key, int ):
                        return getattr(self, key)

                    if key == 0: return self.participation
                    elif key == 1: return self.time
                    elif key == 2: return self.heatto
                    elif key == 3: return self.coolto
                    else: raise IndexError('Numeric index must be an integer 0-3')

                def __len__( self ):
                    return 4

                def __iter__( self ):
                    yield self.participation
                    yield self.time
                    yield self.heatto
                    yield self.coolto

                def __bytes__( self ):
                    cf = self.sched.parent.getCF( self.sched.cf )

                    if self.heatto < -100 or self.heatto > 100:
                        heatto = round(self.heatto)
                    else:
                        heatto = self.heatto
                        # schedule is in C, so convert from F
                        if cf == 'f':
                            heatto = (heatto - 32) / 1.8
                        heatto = round(heatto * 100)
                        heatmod = heatto % 50
                        heatto -= heatmod
                        if heatmod >= 25: heatto += 50

                    if self.coolto < -100 or self.coolto > 100:
                        coolto = round(self.coolto)
                    else:
                        coolto = self.coolto
                        # schedule is in C, so convert from F
                        if cf == 'f':
                            coolto = (coolto - 32) / 1.8
                        coolto = round(coolto * 100)
                        coolmod = coolto % 50
                        coolto -= coolmod
                        if coolmod >= 25: coolto += 50

                    log.info( 'CF is: %r %r %r cool: %r %r %r', cf, self.heatto, heatto / 100, self.coolto, coolto / 100, self.time )

                    # if self.time is a string then it needs to be in 24-hour HH:MM[:SS] format!
                    if isinstance( self.time, str ):
                        tparts = self.time.split( ':' )
                        if len(tparts) >= 2:
                            ptime = (int(tparts[0]) * 60) + int(tparts[1])
                        else:
                            ptime = int(tparts[0])
                    elif isinstance( self.time, int ):
                        ptime = self.time
                    else:
                        ptime = int(self.time)

                    return struct.pack( '>BHhh', self.participation, ptime, heatto, coolto )

                def __repr__( self ):
                    return bytes(self).hex().upper()

            def __init__( self, sched ):
                self.sched = sched
                self.periods = [ ]
                for i in range( 5 ):
                    sp = self.SchedulePeriod( sched )
                    self.periods.append( sp )

            def period_to_idx( self, period ):
                if isinstance( period, int ):
                    if period >= 0 and period < 5:
                        return period
                    raise ValueError('"period" must be an integer in the range 0-4 or a string containing the period name')

                if not isinstance( period, str ):
                    raise ValueError('"period" must be an integer in the range 0-4 or a string containing the period name')

                period = period[0].lower()
                if period == 'w': return 0 # wake
                if period == 'a': return 1 # away
                if period == 'h': return 2 # home
                if period == 's': return 3 # sleep
                if period == 'e': return 4 # extra

                raise ValueError('"period" must be an integer in the range 0-4 or a string containing the period name')

            def __setitem__( self, key, data ):
                if isinstance( key, str ):
                    key = self.period_to_idx( key )
                elif not isinstance( key, int ):
                    setattr( self, key, data )

                if key < 0 or key > 4:
                    raise IndexError('Numeric index must be an integer 0-4')

                self.periods[key] = data

            def __getitem__( self, key ):
                if isinstance( key, str ):
                    key = self.period_to_idx( key )
                elif not isinstance( key, int ):
                    return getattr( self, key )

                if key < 0 or key > 4:
                    raise IndexError('Numeric index must be an integer 0-4')

                return self.periods[key]

            def __len__( self ):
                return 5

            def __iter__( self ):
                for p in self.periods:
                    yield p

            def __bytes__( self ):
                ret = bytearray()
                for period in self.periods:
                    ret += bytearray( bytes( period ) )
                return bytes(ret)

            def __repr__( self ):
                return bytes(self).hex().upper()


        def __init__( self, parent, dps ):
            self.parent = parent
            self.dps = dps
            self.have_data = False
            self.cf = None

            self.day_data = [ ]

            for i in range( 7 ):
                sd = self.ScheduleDay( self )
                self.day_data.append( sd )

        def day_to_idx( self, day ):
            if isinstance( day, int ):
                if day >= 0 and day < 7:
                    return day
                raise ValueError('"day" must be an integer in the range 0-6 or a string containing the day name')

            if not isinstance( day, str ):
                raise ValueError('"day" must be an integer in the range 0-6 or a string containing the day name')

            day = day[:2].lower()
            if day == 'su':   return 0
            if day[0] == 'm': return 1
            if day == 'tu':   return 2
            if day[0] == 'w': return 3
            if day == 'th':   return 4
            if day[0] == 'f': return 5
            if day == 'sa':   return 6

            raise ValueError('"day" must be an integer in the range 0-6 or a string containing the day name')

        def copyDay( self, src, dst ):
            src = self.day_to_idx( src )
            dst = self.day_to_idx( dst )

            for period in range( len(self.day_data[src]) ):
                for itm in range( len(self.day_data[src][period]) ):
                    self.day_data[dst][period][itm] = self.day_data[src][period][itm]

            return self.have_data

        def copyPeriod( self, src_day, src_period, dst_day, dst_period ):
            src_day = self.day_to_idx( src_day )
            #src_period = self.period_to_idx( src_period )
            dst_day = self.day_to_idx( dst_day )
            #dst_period = self.period_to_idx( dst_period )

            for itm in range( len(self.day_data[src_day][src_period]) ):
                self.day_data[dst_day][dst_period][itm] = self.day_data[src_day][src_period][itm]

            return self.have_data

        def setPeriod( self, day, period, **kwargs ):
            day = self.day_to_idx( day )
            #period = self.period_to_idx( period )

            if 'delete' in kwargs:
                self.day_data[day][period] = self.ScheduleDay.SchedulePeriod( self )

            if 'participation' in kwargs:
                self.day_data[day][period].participation = kwargs['participation']

            if 'time' in kwargs:
                self.day_data[day][period].time = kwargs['time']

            if 'heatto' in kwargs:
                self.day_data[day][period].heatto = kwargs['heatto']

            if 'coolto' in kwargs:
                self.day_data[day][period].coolto = kwargs['coolto']

            if self.day_data[day][period][0] > 3 and self.day_data[day][period][1] < 1440:
                if self.day_data[day][period][0] != 0xFF:
                    log.warn('Selected participation flag is out of range, setting to %d', period)
                self.day_data[day][period][0] = period & 3

        def setCF( self, cf ):
            self.cf = cf

        def update( self, data ):
            self.have_data = False

            if len(data) % 7 != 0:
                log.warn( 'Schedule data is in an unknown format, ignoring schedule' )
                return False

            cf = self.parent.getCF( self.cf )
            daylen = int(len(data) / 7)
            for dow in range( 7 ):
                offset = dow * daylen
                day = data[offset:offset+daylen]

                if len(day) % 7 != 0:
                    log.warn( 'Schedule day data for day %d is in an unknown format, ignoring schedule' % dow )
                    return False

                periods = len(day) / 7
                period = -1

                for dayoffset in range( 0, len(day), 7 ):
                    period += 1
                    perioddata = day[dayoffset:dayoffset+7]

                    if len(perioddata) != 7:
                        log.warn( 'Schedule period data for period %d on day %d is in an unknown format, ignoring schedule' % (period, dow) )
                        return False

                    newdata = struct.unpack( '>BHhh', perioddata )

                    for i in range( len(newdata) ):
                        self.day_data[dow][period][i] = newdata[i]

                    # display the time as 24-hour HH:MM
                    if self.day_data[dow][period].time < 1440:
                        hrs = int(self.day_data[dow][period].time / 60)
                        mins = self.day_data[dow][period].time % 60
                        self.day_data[dow][period].time = '%d:%02d' % (hrs,mins)

                    if self.day_data[dow][period].heatto > -10000 and self.day_data[dow][period].heatto < 10000:
                        self.day_data[dow][period].heatto /= 100
                        if cf == 'f':
                            self.day_data[dow][period].heatto = round((self.day_data[dow][period].heatto * 1.8) + 32)

                    if self.day_data[dow][period].coolto > -10000 and self.day_data[dow][period].coolto < 10000:
                        self.day_data[dow][period].coolto /= 100
                        if cf == 'f':
                            self.day_data[dow][period].coolto = round((self.day_data[dow][period].coolto * 1.8) + 32)

            self.have_data = True

        def save( self ):
            return self.parent.set_value( self.dps, self.b64(), nowait=True )

        def __bytes__( self ):
            ret = bytearray()
            for daydata in self.day_data:
                ret += bytearray( bytes( daydata ) )

            return bytes(ret)

        def __repr__( self ):
            #if not self.have_data:
            #    return ''
            return bytes(self).hex().upper()

        def b64(self):
            return base64.b64encode( bytes(self) ).decode('ascii')

        def __iter__(self):
            for d in self.day_data:
                yield d

        def __setitem__( self, key, data ):
            if isinstance( key, str ):
                if key == 'cf':
                    self.cf = data
                    return
                key = self.day_to_idx( key )
            elif not isinstance( key, int ):
                setattr( self, key, data )

            if key < 0 or key > 6:
                raise IndexError('Numeric index must be an integer 0-6')

            self.day_data[key] = data

        def __getitem__( self, key ):
            if isinstance( key, str ):
                key = self.day_to_idx( key )
            elif not isinstance( key, int ):
                return getattr( self, key )

            if key < 0 or key > 6:
                raise IndexError('Numeric index must be an integer 0-6')

            return self.day_data[key]


class ThermostatSensorList(object):
    """
    Represents a list of sensors such as what gets returned in DPS 122

    Args:
        dps: the DPS of this sensor list
        parent_device: the ThermostatDevice which this sensor list is attached to


    The .update(sensordata_list) method parses an update
      Args:
        sensordata_list: either a base64-encoded string such as what DPS 122 contains, or an already-decoded byte string

    The .b64() method returns a base64-encoded string ready for sending
    The str() method returns a hexidecimal string to make it easier to visualize the data

    This class is iterable so you can easily loop through the individual sensors

    I.e.
        ## create a new list
        sensor_list_object = ThermostatSensorList( '122', self ) # for DPS 122

        ## populate the sensor data
        sensor_list_object.update( 'base64 string here' )

        ## send an update after changing a sensor value
        send_dps = { '122': sensor_list_object.b64() }
        payload = d.generate_payload(tinytuya.CONTROL, data)
        d.send(payload)
    """

    def __init__( self, dps, parent_device ):
        self.stated_count = 0
        self.actual_count = 0
        self.sensors = [ ]
        self.parent_device = parent_device

        if isinstance(dps, int):
            dps = str(dps)

        self.dps = dps

    def update(self, sensordata_list):
        changed = [ ]
        if isinstance(sensordata_list, str):
            sensordata_list = base64.b64decode( sensordata_list )
        elif not isinstance(sensordata_list, bytes):
            raise TypeError( 'Unhandled Thermostat Sensor List data type' )

        if( len(sensordata_list) < 1 ):
            self.stated_count = self.actual_count = 0
            self.sensors = [ ]
            return

        lenmod = len(sensordata_list) % 52

        if lenmod == 1:
            self.stated_count = sensordata_list[0]
        elif lenmod == 0:
            self.stated_count = None
        else:
            raise TypeError( 'Unhandled Thermostat Sensor List data length' )

        self.actual_count = int((len(sensordata_list) - lenmod) / 52)

        for i in range( self.actual_count ):
            if i < len(self.sensors):
                if self.sensors[i].parse(sensordata_list[(i*52)+lenmod:((i+1)*52)+lenmod]):
                    changed.append(self.sensors[i])
            else:
                self.sensors.append( self.ThermostatSensorData( self ) )
                self.sensors[i].parse(sensordata_list[(i*52)+lenmod:((i+1)*52)+lenmod])
                # instead of listing every field, just say it was added
                self.sensors[i].changed = [ 'sensor_added' ]
                self.sensors[i].sensor_added = True
                changed.append(self.sensors[i])

            # FIXME should we delete removed sensors?

        return changed

    def __repr__( self ):
        if self.stated_count is not None:
            out = '%02X' % self.stated_count
        else:
            out = ''

        for s in self.sensors:
            out += str(s)

        return out

    def b64(self):
        if self.stated_count is not None:
            b = bytearray( [self.stated_count] )
        else:
            b = bytearray()

        for s in self.sensors:
            b += bytearray( bytes( s ) )
        return base64.b64encode( b ).decode('ascii')

    def __iter__(self):
        for s in self.sensors:
            yield s

    class ThermostatSensorData(object):
        # unpack the 52-byte long binary blob
        struct_format = '>I30s??h?BBBB?h6s'
        keys = ('raw_id', 'raw_name', 'enabled', 'occupied', 'raw_temperature_used', 'online', 'participation', 'battery', 'firmware_version', 'unknown2', 'averaging', 'raw_temperature', 'unknown3')
        raw_temperature_used_idx = keys.index( 'raw_temperature_used' )
        raw_temperature_idx = keys.index( 'raw_temperature' )

        def __init__( self, parent_sensorlist ):
            self.parent_sensorlist = parent_sensorlist
            self.raw_id = 0
            self.raw_name = b'\x00' * 30
            self.name = ''
            self.enabled = True
            self.occupied = True
            self.raw_temperature = 0
            self.temperature = 0.0
            self.raw_temperature_used = 0
            self.temperature_used = 0.0
            self.online = True
            self.participation = 0
            self.battery = 0
            self.unknown2 = 0
            self.firmware_version = 0
            self.averaging = True
            self.unknown3 = b'\x00' * 8
            self.changed = [ ]
            self.want_update = [ ]
            self.sensor_added = True
            self.delay_updates = False


        def parse( self, sensordata ):
            new = struct.unpack( self.struct_format, sensordata )
            self.changed = [ ]
            self.sensor_added = False
            self.delay_updates = False

            if new[self.raw_temperature_idx] == 0:
                new = list(new)
                new[self.raw_temperature_idx] = new[self.raw_temperature_used_idx]
                # "int( N / 50 ) * 50" does a pretty good job of matching what the thermostat does
                new[self.raw_temperature_used_idx] = int(new[self.raw_temperature_used_idx] / 50) * 50

            for i in range(len(self.keys)):
                k = self.keys[i]

                if (k in self.want_update) or (getattr(self, k) != new[i]):
                    self.changed.append( k )
                    setattr(self, k, new[i])

            if 'raw_id' in self.changed:
                self.changed.remove('raw_id')
                self.changed.append('id')
                self.id = '%08x' % self.raw_id

            if 'raw_name' in self.changed:
                self.changed.remove('raw_name')
                self.changed.append('name')
                self.name = self.raw_name.strip(b'\x00').decode('utf8')

            if 'raw_temperature' in self.changed:
                self.changed.remove('raw_temperature')
                self.changed.append('temperature')
                self.temperature = self.raw_temperature / 100.0

            if 'raw_temperature_used' in self.changed:
                self.changed.remove('raw_temperature_used')
                self.changed.append('temperature_used')
                self.temperature_used = self.raw_temperature_used / 100.0

            self.want_update = [ ]

            return (len(self.changed) != 0)

        def delayUpdates( self ):
            self.delay_updates = True

        def setName( self, name ):
            self.name = name
            # the app limits the length to 20 chars, so lets do the same
            self.raw_name = name[:20].encode('utf8').rjust( 30, b'\0' )
            self.want_update.append( 'raw_name' )
            self.sendUpdates(False)

        def setEnabled( self, ena ):
            self.enabled = ena
            self.want_update.append( 'enabled' )
            self.sendUpdates(False)

        def setOccupied( self, occ ):
            self.occupied = occ
            self.want_update.append( 'occupied' )
            self.sendUpdates(False)

        def setParticipation( self, flag, val=True ):
            self.want_update.append( 'participation' )

            if isinstance( flag, str ):
                mask = 1 << ( 'wake', 'away', 'home', 'sleep' ).index( flag )
                if val:
                    self.participation |= mask
                else:
                    self.participation &= ~mask
            elif isinstance( flag, int ):
                self.participation = flag

            self.sendUpdates(False)

        #def clearParticipation( self, flag ):
        #    return self.setParticipation( 0, False )

        def getParticipation( self, flag ):
            if isinstance( flag, str ):
                mask = 1 << ( 'wake', 'away', 'home', 'sleep' ).index( flag )
                if (self.participation & mask) == mask:
                    return True
                return False
            elif isinstance( flag, int ):
                if (self.participation & flag) == flag:
                    return True
                return False

            return False

        ## technically the battery level and firmware version can also be changed, no idea why someone would want to do that though
        #def setBattery( self, new_battery ):
        #def setFirmware( self, new_firmware ):

        def setUnknown2( self, u2 ):
            self.unknown2 = u2
            self.want_update.append( 'unknown2' )
            self.sendUpdates(False)

        def setUnknown3( self, u3 ):
            if not isinstance( u3, bytes ):
                u3 = bytes( u3 )

            if len( u3 ) < 8:
                u3 = u3 + (b'\x00' * (8 - len( u3 )))

            self.unknown3 = u3[:8]
            self.want_update.append( 'unknown3' )
            self.sendUpdates(False)

        def sendUpdates( self, force=True ):
            if (not force) and self.delay_updates:
                return

            self.delay_updates = False
            idx = self.parent_sensorlist.parent_device.sensor_dps.index( self.parent_sensorlist.dps )
            self.parent_sensorlist.parent_device.set_value( self.parent_sensorlist.dps, self.parent_sensorlist.b64(), nowait=True )

        def __repr__( self ):
            return bytes(self).hex().upper()

        def __bytes__( self ):
            try:
                return struct.pack( self.struct_format, *(getattr(self, k) for k in self.keys) )
            except:
                log.exception( 'Error while attempting to pack %s with %r/%r/%r/%r/%r/%r/%r/%r/%r/%r/%r/%r', self.struct_format, *(getattr(self, k) for k in self.keys) )
                raise




# TinyTuya Contrib IRRemoteControlDevice Module
# -*- coding: utf-8 -*-
"""
 A community-contributed Python module to add support for Tuya WiFi smart universal remote control simulators

 This module attempts to provide everything needed so there is no need to import the base tinytuya module

 Module Author: Alexey 'Cluster' Avdyukhin (https://github.com/clusterm)
 Rewritten by uzlonewolf (https://github.com/uzlonewolf) for new devices and IR format conversion

 Local Control Classes
    IRRemoteControlDevice(..., version=3.3)
        This class uses a default version of 3.3
        See OutletDevice() for the other constructor arguments

    Functions:
        ir = IRRemoteControlDevice(..., control_type=None)
            -> will immediately connect to the device to try and detect the control type if control_type is not provided
               control_type=1 for older devices using DPS 201/202
               control_type=2 for newer devices using DPS 1-13

        ir.detect_control_type()
            -> polls device status to try and detect the control type

        ir.send_command( mode, data={} )
            -> sends a command to the device
               when mode is 'send', data is parsed for the data to send
                   data = { "base64_code": "..." } or
                   data = { "head": "...", "key": "..." }
               all other commands are sent though as-is without data

        ir.study_start()
        ir.study_end()
            -> start or end a study session

        ir.receive_button( timeout )
            -> call this method and press button on real remote control to read its code in Base64 format
               timeout - maximum time to wait for button press

        ir.send_button( base64_code )
            -> simulate a learned (raw base64-encoded) button press

        ir.send_key( head, key )
            -> send a head/key pair

        ir.build_head( freq=38, bit_time=0, zero_time=0, one_time=0, bit_time_type=1, timings=[], convert_time=True )
            -> build a 'head' section
               'freq' is in kHz
               if bit_time, zero_time, or one_time evaluate to False, timings are taken from timings[] as needed
               if convert_time is True, timings are in microseconds and converted as needed.
                   when False, timings are sent as-is

        IRRemoteControlDevice.print_pulses ( pulses )
            -> pretty-print a sequence of pulses and gaps length

        IRRemoteControlDevice.base64_to_pulses ( code_base_64 )
            -> convert Base64-encoded button code to sequence of pulses and gaps length

        IRRemoteControlDevice.pulses_to_base64 ( pulses )
            -> convert sequence of pulses and gaps length to Base64-encoded button code

        IRRemoteControlDevice.head_key_to_pulses ( head='...', key='...' )
            -> convert head/key pair to sequence of pulses and gaps
               'head' can be None when the key is raw bytes in base64
               'key' must begin with '00' through 'FF' when it is not raw bytes in base64

        IRRemoteControlDevice.pulses_to_head_key ( pulses, fudge=0.1, freq=38 )
            -> attempts to pack a sequence of pulses and gaps into a head/key pair
               pulses/gaps within 10% (fudge=0.1) are assumed to be the same and are merged together

        IRRemoteControlDevice.hex_to_pulses ( code_hex )
            -> convert HEX-encoded button code to sequence of pulses and gaps length
               HEX-encoded codes are used in the Cloud API

        IRRemoteControlDevice.pulses_to_hex ( pulses )
            -> convert sequence of pulses and gaps length to HEX-encoded button code
               HEX-encoded codes are used in the Cloud API

        IRRemoteControlDevice.nec_to_pulses ( address, data=None )
            -> convert a 32-bit NEC button code (when data=None) or a 8/16-bit address and 8-bit data to sequence of pulses and gaps length
               address - a 32-bit NEC button code (when data=None), or a 8-bit or 16-bit address
               data - 8-bit data when address is 8-bit or 16-bit

        IRRemoteControlDevice.pulses_to_nec ( pulses )
            -> convert sequence of pulses and gaps length to a NEC button code
               returns an array of dicts containing 'type'="nec", 'address' and 'data' (if valid), 'uint32' raw data, and 'hex' hex-encoded data

        IRRemoteControlDevice.samsung_to_pulses ( address, data=None )
            -> similar to nec_to_pulses() but for the Samsung format (start pulse 4.5ms instead of 9ms)

        IRRemoteControlDevice.pulses_to_samsung ( pulses )
            -> similar to pulses_to_nec() but for the Samsung format
               returns same array of dict as pulses_to_nec() but with 'type'="samsung"

        IRRemoteControlDevice.pronto_to_pulses ( pronto )
            -> convert a Pronto code string to sequence of pulses and gaps length

        IRRemoteControlDevice.pulses_to_pronto ( pulses )
            -> convert sequence of pulses and gaps length to a Pronto code string

        IRRemoteControlDevice.width_encoded_to_pulses ( uint32, start_mark=9000, start_space=4500, pulse_one=563, pulse_zero=563,
          space_one=1688, space_zero=563, trailing_pulse=563, trailing_space=30000 )
            -> flexible uint32 to sequence of pulses encoder
               default values are for NEC format, set start_mark=4500 for Samsung format

        IRRemoteControlDevice.pulses_to_width_encoded ( pulses, start_mark=None, start_space=None, pulse_threshold=None, space_threshold=None )
            -> flexible space-width or pulse-width to uint32 decoder
               converts sequence of pulses to a 32-bit unsigned integer
               recommended *_threshold is `(length_of_zero + length_of_one) / 2`

"""

import base64
import json
import logging
import struct
import time

from ..core import Device, log, CONTROL

class IRRemoteControlDevice(Device):
    CMD_SEND_KEY_CODE = "send_ir"   # Command to start sending a key
    DP_SEND_IR        = "201"       # ir_send, send and report (read-write)
    DP_LEARNED_ID     = "202"       # ir_study_code, report only (read-only)
    DP_MODE           =   "1"
    DP_LEARNED_REPORT =   "2"
    DP_HEAD           =   "3"
    DP_KEY_CODE       =   "4"
    DP_KEY_CODE2      =   "5"
    DP_KEY_CODE3      =   "6"
    DP_KEY_CODE4      =  "11"
    DP_KEY_STUDY      =   "7"
    DP_KEY_STUDY2     =   "8"
    DP_KEY_STUDY3     =   "9"
    DP_KEY_STUDY4     =  "12"
    DP_SEND_DELAY     =  "10"
    DP_CODE_TYPE      =  "13"
    NSDP_CONTROL = "control"       # The control commands
    NSDP_STUDY_CODE = "study_code" # Report learned IR codes
    NSDP_IR_CODE = "ir_code"       # IR signal decoding2
    NSDP_KEY_CODE = "key_code"     # Remote key code
    NSDP_KEY_CODE2 = "key_code2"   # Remote key code 2
    NSDP_KEY_CODE3 = "key_code3"   # Remote key code 3
    NSDP_KEY_CODE4 = "key_code4"   # Remote key code 4
    NSDP_KEY_STUDY = "key_study"   # Send the learning code 1
    NSDP_KEY_STUDY2 = "key_study2" # Send the learning code 2
    NSDP_KEY_STUDY3 = "key_study3" # Send the learning code 3
    NSDP_KEY_STUDY4 = "key_study4" # Send the learning code 4
    NSDP_DELAY_TIME = "delay_time" # IR code transmission delay
    NSDP_TYPE = "type"             # The identifier of an IR library
    NSDP_DELAY = "delay"           # Actually used but not documented
    NSDP_HEAD = "head"             # Actually used but not documented
    NSDP_KEY1 = "key1"             # Actually used but not documented
    KEY1_SYMBOL_LIST = "@#$%^&*()QWRLTXKVNM{}[]JUP<>|=HS~" # Timing symbols used in key1

    def __init__(self, *args, **kwargs):
        # set the default version to 3.3 as there are no 3.1 devices
        if 'version' not in kwargs or not kwargs['version']:
            kwargs['version'] = 3.3

        control_type = 0
        if 'control_type' in kwargs:
            control_type = kwargs['control_type']
            del kwargs['control_type']

        super(IRRemoteControlDevice, self).__init__(*args, **kwargs)

        self.disabledetect = True
        self.control_type = control_type
        if not self.control_type:
            self.detect_control_type()

    def detect_control_type( self ):
        # This is more difficult than it seems.  Neither device responds to status() after
        #   a reboot until after a command is sent.  201 devices do not respond to study_end
        #   if they are already in that mode.
        old_timeout = self.connection_timeout
        old_persist = self.socketPersistent
        self.set_socketTimeout( 1 )
        self.set_socketPersistent( True )
        self.control_type = 1
        self.study_end()
        self.control_type = 2
        self.study_end()
        self.control_type = 0
        status = self.status()
        while status and 'dps' in status:
            # original devices using DPS 201/202
            if self.DP_SEND_IR in status['dps']:
                log.debug( 'Detected control type 1' )
                self.control_type = 1
            # newer devices using DPS 1-13
            elif self.DP_MODE in status['dps']:
                log.debug( 'Detected control type 2' )
                self.control_type = 2
            status = self._send_receive(None)
        if not self.control_type:
            log.warning( 'Detect control type failed! control_type= must be set manually' )
        self.set_socketTimeout( old_timeout )
        self.set_socketPersistent( old_persist )

    def send_command( self, mode, data={} ):
        if mode == 'send':
            if self.control_type == 1:
                command = {
                    IRRemoteControlDevice.NSDP_CONTROL: "send_ir",
                    IRRemoteControlDevice.NSDP_TYPE: 0,
                }

                if 'base64_code' in data:
                    command[IRRemoteControlDevice.NSDP_HEAD] = ''
                    command[IRRemoteControlDevice.NSDP_KEY1] = '1' + data['base64_code']
                elif 'head' in data and 'key' in data:
                    command[IRRemoteControlDevice.NSDP_HEAD] = data['head']
                    command[IRRemoteControlDevice.NSDP_KEY1] = '0' + data['key']
                self.set_value( IRRemoteControlDevice.DP_SEND_IR, json.dumps(command), nowait=True )
            elif self.control_type == 2:
                mode = 'study_key' if 'base64_code' in data else 'send_ir'
                command = {
                    IRRemoteControlDevice.DP_MODE: mode,
                    IRRemoteControlDevice.DP_CODE_TYPE: 0,
                }
                if 'base64_code' in data:
                    command[IRRemoteControlDevice.DP_KEY_STUDY] = data['base64_code']
                elif 'head' in data and 'key' in data:
                    command[IRRemoteControlDevice.DP_HEAD] = data['head']
                    command[IRRemoteControlDevice.DP_KEY_CODE] = data['key']
                self.set_multiple_values( command, nowait=True )
        elif self.control_type == 1:
            command = { IRRemoteControlDevice.NSDP_CONTROL: mode }
            self.set_value( IRRemoteControlDevice.DP_SEND_IR, json.dumps(command), nowait=True )
        elif self.control_type == 2:
            self.set_value( IRRemoteControlDevice.DP_MODE, mode, nowait=True )

    def study_start( self ):
        self.send_command( 'study' )

    def study_end( self ):
        self.send_command( 'study_exit' )

    def receive_button( self, timeout=30 ):
        log.debug("Receiving button")
        # Exit study mode in case it's enabled
        self.study_end()
        # Enable study mode
        self.study_start()

        # Receiving button code
        response = None
        response_code = None
        found = False
        # Remember old timeout and set new timeout
        old_timeout = self.connection_timeout
        end_at_time = time.time() + timeout
        try:
            while end_at_time > time.time():
                timeo = round(time.time() - end_at_time)
                if timeo < 1: timeo = 1
                self.set_socketTimeout(timeo)

                log.debug("Waiting for button...")
                response = self._send_receive(None)
                if response == None:
                    # Nothing received
                    log.debug("Timeout")
                elif type(response) != dict or "dps" not in response:
                    # Some unexpected result
                    log.debug("Unexpected response: %r", response)
                    response_code = response # Some error message? Pass it.
                    break
                elif self.DP_LEARNED_ID in response["dps"]:
                    # Button code received, extracting it as Base64 string
                    response_code = response["dps"][self.DP_LEARNED_ID]
                    found = True
                    break
                elif self.DP_LEARNED_REPORT in response["dps"]:
                    response_code = response["dps"][self.DP_LEARNED_REPORT]
                    found = True
                    break
                else:
                    # Unknown DPS
                    log.debug("Unknown DPS in response: %r", response)
                    response_code = response # Pass it if we do not get a response we like
                    # try again
        finally:
            # Revert timeout
            self.set_socketTimeout(old_timeout)

        if found:
            self.print_pulses( response_code )

        # Exit study mode
        self.study_end()

        return response_code

    def send_button( self, base64_code ):
        log.debug( 'Sending Learned Button: ' + base64_code)
        self.print_pulses( base64_code )
        return self.send_command( 'send', {'base64_code': base64_code} )

    def send_key( self, head, key ):
        log.debug( 'Sending Key: %r / %r', head, key )
        return self.send_command( 'send', { 'head': head, 'key': key } )

    @staticmethod
    def build_head( freq=38, bit_time=0, zero_time=0, one_time=0, bit_time_type=1, timings=[], convert_time=True ):
        timings = list(timings)
        freq = round( freq * 100)
        if not bit_time and len(timings) > 0:
            bit_time = timings[0]
            timings = timings[1:]
        if not zero_time and len(timings) > 0:
            zero_time = timings[0]
            timings = timings[1:]
        if not one_time and len(timings) > 0:
            one_time = timings[0]
            timings = timings[1:]

        if convert_time:
            time_base = 100000.0 / freq
            bit_time = round( bit_time / time_base )
            zero_time = round( zero_time / time_base )
            one_time = round( one_time / time_base )
            for i in range(len(timings)):
                timings[i] = round( timings[i] / time_base )

        head = '%02X%04X0000000000' % (bit_time_type, freq)
        head += '%02X%04X%04X%04X' % (len(timings) + 3, bit_time, zero_time, one_time)
        for i in timings:
            head += '%04X' % i

        return head

    @staticmethod
    def print_pulses( base64_code, use_log=None ):
        if not use_log: use_log = log
        if type(base64_code) == list:
            pulses = base64_code
        else:
            pulses = IRRemoteControlDevice.base64_to_pulses(base64_code)
        message = "Pulses and gaps (microseconds): " + ' '.join([f'{"p" if i % 2 == 0 else "g"}{pulses[i]}' for i in range(len(pulses))])
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug( message )
        return message

    @staticmethod
    def base64_to_pulses( code_base_64 ):
        if len(code_base_64) % 4 == 1 and code_base_64.startswith("1"):
            # code can be padded with "1"
            code_base_64 = code_base_64[1:]
        raw_bytes = base64.b64decode(code_base_64)
        fmt = '<%dH' % (len(raw_bytes) >> 1)
        return list(struct.unpack(fmt, raw_bytes))

    @staticmethod
    def pulses_to_base64( pulses ):
        fmt = '<' + str(len(pulses)) + 'H'
        return base64.b64encode( struct.pack(fmt, *pulses) ).decode("ascii")

    @staticmethod
    def head_key_to_pulses( head, key ):
        if len(key) < 4:
            raise ValueError( '"key" must be at least 4 characters' )

        if not head:
            return IRRemoteControlDevice.base64_to_pulses( key )

        if len(head) < 18:
            raise ValueError( '"head" must be at least 18 characters' )

        head = bytearray.fromhex( head )
        headtype, timescale, unused1, unused2, num_timings = struct.unpack( '>BHHHH', head[:9] )
        headlen = num_timings * 2 # times are 16-bit
        timebase = 100000.0 / timescale
        symbols = IRRemoteControlDevice.KEY1_SYMBOL_LIST[:num_timings]
        try:
            repeat = int( key[:2], 16 )
        except:
            raise ValueError( 'First 2 digit of "key" must be a hexidecimal byte' )
        key = key[2:]

        # 'head' type 1 uses '@' for 0 and '#' for 1
        # 'head' type 2 uses '#' for 0 and '$' for 1

        if headtype == 1:
            bit_timimgs = ( '@@', '@#' )
        elif headtype == 2:
            bit_timimgs = ( '@#', '@$' )
        else:
            raise ValueError( 'Unhandled "head" type: %d' % headtype )

        if len(head) != (headlen+9):
            raise ValueError( '"head" must be %d characters' % ((headlen+9)*2) )

        # unpack the timing values, however many there are
        fmt = '>%dH' % num_timings
        timings = struct.unpack( fmt, head[9:] )
        symbol_timings = {}

        for i in range(num_timings):
            symbol_timings[symbols[i]] = round(timebase * timings[i])

        if False:
            print( 'Head:' )
            print( 'Frequency: %r kHz, Time Base: %r' % (timescale / 100.0, timebase) )
            print( 'Bit Symbols: 0 = %s, 1 = %s' % bit_timimgs)
            print( 'Symbol Timings:' )
            for i in range(num_timings):
                print( '  %s = %r microseconds' % (symbols[i], symbol_timings[symbols[i]]) )
            print( '' )
            print( 'Key:' )
            print( 'Send count:', repeat )
            print( 'Code:', key )

        # although it's not as effiecient, it's easier to see what's going on if
        #  you first unpack the packed bits into their symbol pairs, and then
        #  expand those symbols to their timing times

        expanded = ''
        while key:
            cnt = 0
            # first, copy symbols as-is
            for c in key:
                if c not in symbols:
                    break

                expanded += c
                cnt += 1

            key = key[cnt:]
            if not key:
                # all finished
                break

            # next, expand packed bits
            #print( 'Unpacking:', key[:4] )
            byts, bits = struct.unpack( '>BB', bytearray.fromhex( key[:4] ) )
            key = key[4:]
            if byts != 0:
                # if the first byte is not 0, read in and transmit bytes until a symbol is encountered
                cnt = 0
                bits = 0
                data = ''
                for c in key:
                    c2 = ord(c.upper())
                    if c2 < 0x30 or c2 > 0x46 or (c2 > 0x39 and c2 < 0x41):
                        # it's a symbol, we're done
                        break
                    data += c
                    cnt += 1
                    bits += 4
                if (len(data) % 2):
                    data += '0'
            else:
                # if the first byte is 0, the next byte is how many bits to transmit
                byts = int( (bits + 7) / 8 )
                cnt = byts * 2
                data = key[:cnt]

            key = key[cnt:]
            byts = bytearray.fromhex( data )
            # unpack the bits into symbol pairs
            while bits:
                d = byts[0]
                byts = byts[1:]
                for i in range(8):
                    if not bits:
                        break
                    # devices transmit MSB first
                    if (d & 0x80) == 0x80:
                        expanded += bit_timimgs[1]
                    else:
                        expanded += bit_timimgs[0]
                    d <<= 1
                    bits -= 1

        if False:
            print( 'Expanded Code:', expanded )
            print( '' )
            print( 'Pulse train:' )

            # expand the symbols into their pulses
            for c in expanded:
                print( symbol_timings[c], end=' ' )
            print( '' )

        return [symbol_timings[c] for c in expanded]


    @staticmethod
    def pulses_to_head_key( pulses, fudge=0.1, freq=38 ):
        mylog = log.getChild( 'pulses_to_head_key' )

        if len(pulses) < 2:
            return None

        if len(pulses) % 2 == 1:
            pulses = list(pulses)
            pulses.append(pulses[0])

        ps_count = { }
        for current_ps_time in pulses:
            if current_ps_time not in ps_count:
                ps_count[current_ps_time] = 1
            else:
                ps_count[current_ps_time] += 1
        ps_map = IRRemoteControlDevice._merge_similar_pulse_times( ps_count, fudge )

        # should we process the pulses and spaces separately?
        # combining them seems to give good results
        if False:
            p_count = { }
            s_count = { }
            is_pulse = False
            for current_ps_time in pulses:
                is_pulse = not is_pulse

                if current_ps_time in ps_map:
                    current_ps_time = ps_map[current_ps_time]

                if is_pulse:
                    if current_ps_time not in p_count:
                        p_count[current_ps_time] = 1
                    else:
                        p_count[current_ps_time] += 1
                else:
                    if current_ps_time not in s_count:
                        s_count[current_ps_time] = 1
                    else:
                        s_count[current_ps_time] += 1

            mylog.debug( 'p_count: %r, s_count: %r', p_count, s_count )
            p_map = IRRemoteControlDevice._merge_similar_pulse_times( p_count, fudge )
            s_map = IRRemoteControlDevice._merge_similar_pulse_times( s_count, fudge )
            mylog.debug('merged pulse map: %r', p_map)
            mylog.debug('merged space map: %r', s_map)
        else:
            p_map = s_map = ps_map
            mylog.debug('merged pulse+space map: %r', ps_map)

        # convert the list of pulse and space lengths into a string to
        #  make it easier to group and count unique sequences.
        # the first unique pulse will get the symbol 'A' while
        #  the first space becomes 'a'.  The next is 'B' and 'b'.
        # all pulses/spaces of the same length get the same letter
        # I.e. [ 4523 4523 552 1683 552 1683 552 552 552 552 ] becomes
        #        A    a    B   b    B   b    B   c   B   c   -> AaBbBbBcBc
        # we can then substring count 'Bb' and 'Bc'
        symbol_pattern = ''
        symbol_list = { }
        p_key_map = { }
        s_key_map = { }
        is_pulse = False
        for current_ps_time in pulses:
            is_pulse = not is_pulse
            if is_pulse:
                #if this length was combined, use the new (averaged) value
                k = p_map[current_ps_time] if current_ps_time in p_map else current_ps_time

                # if this length has not been seen yet, assign it a letter
                if k not in p_key_map:
                    next_letter = chr(len(p_key_map) + 0x41) # A-Z
                    #mylog.debug('adding pulse %r %r', k, next_letter)
                    p_key_map[k] = { 'count': 1, 'char': next_letter }
                    if next_letter not in symbol_list:
                        symbol_list[next_letter] = [k, False]
                else:
                    p_key_map[k]['count'] += 1
                symbol_pattern += p_key_map[k]['char']
            else:
                #if this length was combined, use the new (averaged) value
                k = s_map[current_ps_time] if current_ps_time in s_map else current_ps_time

                # if this length has not been seen yet, assign it a letter
                if k not in s_key_map:
                    next_letter = chr(len(s_key_map) + 0x61) # a-z
                    #mylog.debug('adding SPACE %r %r', k, next_letter)
                    s_key_map[k] = { 'count': 1, 'char': next_letter }
                    if next_letter not in symbol_list:
                        symbol_list[next_letter] = [k, False]
                else:
                    s_key_map[k]['count'] += 1
                symbol_pattern += s_key_map[k]['char']

        mylog.debug( 'symbol pattern: %r', symbol_pattern )
        mylog.debug( 'symbol list: %r', symbol_list )

        # find the most-commonly-ocurring pulse and space lengths
        pmax = { 'count': 0, 'time': 0 }
        smax = { 'count': 0, 'time': 0 }
        for k in p_key_map:
            if p_key_map[k]['count'] > pmax['count']:
                pmax['count'] = p_key_map[k]['count']
                pmax['time'] = k
        for k in s_key_map:
            if s_key_map[k]['count'] > smax['count']:
                smax['count'] = s_key_map[k]['count']
                smax['time'] = k

        k = smax['time']
        space_letter = s_key_map[k]['char']

        k = pmax['time']
        pulse_letter = p_key_map[k]['char']

        mylog.debug( 'most common space: %r %r', space_letter, smax )
        mylog.debug( 'most common pulse: %r %r', pulse_letter, pmax )

        encoding_type_shortest = [None, None]
        encoding_type_symbol_list = [{}, {}]
        bit_time_types = [0, 0]
        # calculate the head and key for both space-width and pulse-width encoding
        # we will use the shorter of the 2 as the final head/key
        for encoding_type in range( 2 ):
            mylog.debug( '' )
            current_letter = pulse_letter if encoding_type == 0 else space_letter
            encoding_type_name = 'pulse' if not encoding_type else 'space'
            mylog.debug( 'Trying encoding_type %r - character %r', encoding_type_name, current_letter )
            pat_counts = {}
            for i in range( encoding_type, len(symbol_pattern), 2 ):
                if symbol_pattern[i] == current_letter:
                    k = symbol_pattern[i:i+2]
                    if len(k) == 2:
                        if k not in pat_counts:
                            pat_counts[k] = 1
                        else:
                            pat_counts[k] += 1

            mylog.debug( 'pat_counts: %r', pat_counts )

            # find the most-common and next-most-common pattern pair
            pat_max = [0, '']
            pat_next_max = [0, '']
            for k in pat_counts:
                if pat_counts[k] > pat_max[0]:
                    pat_max[0] = pat_counts[k]
                    pat_max[1] = k
            for k in pat_counts:
                if k != pat_max[1] and pat_counts[k] > pat_next_max[0]:
                    pat_next_max[0] = pat_counts[k]
                    pat_next_max[1] = k

            mylog.debug( 'pat_max: %r, pat_next_max: %r', pat_max, pat_next_max)

            # reset from the previous 'encoding_type' loop
            for k in symbol_list:
                symbol_list[k][1] = False

            try_bitfield = True
            bit_symbol_pattern = ''
            full_symbol_pattern = symbol_pattern

            if pat_max[0] and not pat_next_max[0]:
                a = pat_max[1][0]
                symbol_list[a][1] = '@'
                a2 = pat_max[1][1]
                symbol_list[a2][1] = '#'
                zero_symbol = pat_max[1]
                one_symbol = 'DEADBEEF'
            # assign timing symbols to the most-common and next-most-common lengths
            elif pat_max[0] and pat_next_max[0]:
                a = pat_max[1][0]
                b = pat_next_max[1][0]
                if symbol_list[a][0] == symbol_list[b][0]:
                    # pulses are the same, it might be space-width encoded
                    symbol_list[a][1] = symbol_list[b][1] = '@'
                    a2 = pat_max[1][1]
                    b2 = pat_next_max[1][1]
                    if symbol_list[a2][0] < symbol_list[b2][0]:
                        symbol_list[a2][1] = '#'
                        symbol_list[b2][1] = '$'
                        zero_symbol = pat_max[1]
                        one_symbol = pat_next_max[1]
                    else:
                        symbol_list[a2][1] = '$'
                        symbol_list[b2][1] = '#'
                        zero_symbol = pat_next_max[1]
                        one_symbol = pat_max[1]
                else:
                    # pulses are not the same
                    if symbol_list[a][0] < symbol_list[b][0]:
                        symbol_list[a][1] = '#'
                        symbol_list[b][1] = '$'
                        zero_symbol = pat_max[1]
                        one_symbol = pat_next_max[1]
                    else:
                        symbol_list[a][1] = '$'
                        symbol_list[b][1] = '#'
                        zero_symbol = pat_next_max[1]
                        one_symbol = pat_max[1]

                    a = pat_max[1][1]
                    b = pat_next_max[1][1]
                    if symbol_list[a][0] == symbol_list[b][0]:
                        # but all spaces are the same, probably pulse-width encoded
                        symbol_list[a][1] = symbol_list[b][1] = '@'
                        bit_symbol_pattern = symbol_pattern[0]
                        full_symbol_pattern = symbol_pattern[1:]
                    else:
                        symbol_list[a][1] = '@'
                        try_bitfield = False

            mylog.debug('initial symbol list: %r', symbol_list)

            # if the common length and the zero length are the same, combine them as 'head type 1'
            # first find the symbols for '@' and '#'
            bit_start_symbol = bit_zero_symbol = None
            for k in symbol_list:
                if symbol_list[k][1] == '@':
                    bit_start_symbol = k
                elif symbol_list[k][1] == '#':
                    bit_zero_symbol = k

            # they are the same, combine them into 'head type 1'
            if bit_start_symbol and bit_zero_symbol and symbol_list[bit_start_symbol][0] == symbol_list[bit_zero_symbol][0]:
                bit_time_types[encoding_type] = 1
                symbols_available = list(IRRemoteControlDevice.KEY1_SYMBOL_LIST[2:])
                symbol_list[bit_zero_symbol][1] = '@'
                for k in symbol_list:
                    if symbol_list[k][1] == '$':
                        symbol_list[k][1] = '#'
            # they are different, use 'head type 2'
            else:
                bit_time_types[encoding_type] = 2
                symbols_available = list(IRRemoteControlDevice.KEY1_SYMBOL_LIST[2:])

            # start assigning times to symbols
            # the common/0/1 symbols were already set above
            time_symbols = { }
            for k in symbol_list:
                if symbol_list[k][1]:
                    c = symbol_list[k][0]
                    time_symbols[c] = symbol_list[k][1]

            # assign symbols to the remaining pulse/space times
            need_abort = False
            mylog.debug('symbol list before assignment: %r', symbol_list )
            for k in symbol_list:
                if not symbol_list[k][1]:
                    t = symbol_list[k][0]
                    if t in time_symbols:
                        symbol_list[k][1] = time_symbols[t]
                        continue
                    if not symbols_available:
                        #raise ValueError( 'Cannot convert pulses to head/key, too many unique pulse/space values' )
                        mylog.debug( 'Cannot convert pulses to head/key, too many unique pulse/space values!' )
                        #return None
                        need_abort = True
                        break
                    s = symbols_available.pop( 0 )
                    symbol_list[k][1] = s
                    time_symbols[t] = s
            mylog.debug('symbol list after assignment:  %r', symbol_list )
            mylog.debug('unique time symbols: %r', time_symbols)
            if need_abort:
                mylog.debug('!! need_abort !!')
                continue

            mylog.debug( 'zero sequence: %r, one sequence: %r', zero_symbol, one_symbol ) # pylint: disable=used-before-assignment


            raw_symbol_pattern = ''
            for c in symbol_pattern:
                raw_symbol_pattern += symbol_list[c][1]
            mylog.debug( 'raw symbol pattern: %r', raw_symbol_pattern )

            # see if we can condense bitfields into len+data
            if try_bitfield:
                if encoding_type:
                    c = full_symbol_pattern[0]
                    bit_symbol_pattern += symbol_list[c][1]

                bits = data = 0
                byts = []
                removed = ''
                # the len+2 is to make sure we catch any trailing bits
                for i in range( encoding_type, len(full_symbol_pattern)+2, 2 ):
                    k = full_symbol_pattern[i:i+2]
                    k_symbol_pattern = ''
                    for c in k:
                        k_symbol_pattern += symbol_list[c][1]
                    if k == zero_symbol:
                        removed += k_symbol_pattern
                        bits += 1
                        if bits == 8:
                            byts.append( data )
                            bits = data = 0
                    elif k == one_symbol:
                        removed += k_symbol_pattern
                        bits += 1
                        data |= 1 << (8 - bits)
                        if bits == 8:
                            byts.append( data )
                            bits = data = 0
                    else:
                        if bits or byts:
                            new_bitfield = IRRemoteControlDevice._build_key_bitfield( bits, data, byts )
                            # if the new bitfield is longer than the original timing symbols, don't use it
                            if len(new_bitfield) < len(removed):
                                bit_symbol_pattern += new_bitfield
                            else:
                                bit_symbol_pattern += removed
                        bits = data = 0
                        byts = []
                        removed = ''
                        bit_symbol_pattern += k_symbol_pattern
                mylog.debug( 'bit symbol pattern: %r', bit_symbol_pattern )

                # this should not be needed due to the check in the loop above, but make sure anyway
                if len(bit_symbol_pattern) > len(raw_symbol_pattern):
                    mylog.debug( 'Bitfield pattern is longer than pulse/space symbol lists, using shorter symbol list' )
                    bit_symbol_pattern = raw_symbol_pattern
            else:
                mylog.debug( 'Not attempting bitfield' )
                bit_symbol_pattern = raw_symbol_pattern

            # save the results to see which one is better
            encoding_type_shortest[encoding_type] = bit_symbol_pattern
            encoding_type_symbol_list[encoding_type] = { }
            for k in symbol_list:
                j = symbol_list[k][1]
                encoding_type_symbol_list[encoding_type][j] = symbol_list[k][0]

        mylog.debug( '' )

        if not encoding_type_shortest[0]:
            key1 = encoding_type_shortest[1]
            new_symbol_list = encoding_type_symbol_list[1]
            bit_time_type = bit_time_types[1]
        elif not encoding_type_shortest[1]:
            key1 = encoding_type_shortest[0]
            new_symbol_list = encoding_type_symbol_list[0]
            bit_time_type = bit_time_types[0]
        elif len(encoding_type_shortest[0]) <= len(encoding_type_shortest[1]):
            key1 = encoding_type_shortest[0]
            new_symbol_list = encoding_type_symbol_list[0]
            bit_time_type = bit_time_types[0]
        else:
            key1 = encoding_type_shortest[1]
            new_symbol_list = encoding_type_symbol_list[1]
            bit_time_type = bit_time_types[1]
        #mylog.debug(new_symbol_list)

        # copy over the symbol times, making sure they're in the correct order
        time_symbols = []
        for c in IRRemoteControlDevice.KEY1_SYMBOL_LIST:
            if c in new_symbol_list:
                time_symbols.append( new_symbol_list[c] )
            elif len(time_symbols) < 3:
                time_symbols.append( 100 )
            else:
                break

        header = IRRemoteControlDevice.build_head( freq=freq, bit_time_type=bit_time_type, timings=time_symbols )

        mylog.debug( 'Space-Width Encoded: %r Timings: %r', encoding_type_shortest[0], encoding_type_symbol_list[0] )
        mylog.debug( 'Pulse-Width Encoded: %r Timings: %r', encoding_type_shortest[1], encoding_type_symbol_list[1] )
        mylog.debug( 'new pattern: %r / %r', header, key1 )
        mylog.debug( p_key_map )
        mylog.debug( s_key_map )
        #mylog.debug( symbol_pattern )
        return header, '01' + key1

    @staticmethod
    def _merge_similar_pulse_times( p_count, fudge ):
        p_map = { }
        mod = True
        while mod:
            mod = False
            merge = None
            for current_ps_time in p_count:
                pfudge = current_ps_time * fudge
                pmin = current_ps_time - pfudge
                pmax = current_ps_time + pfudge
                for p_check in p_count:
                    if current_ps_time == p_check:
                        continue
                    if p_check >= pmin and p_check <= pmax:
                        merge = (current_ps_time, p_check)
                        #print( 'merging', merge )
                        break
                    #else:
                    #    print('not merging, pmin < p_check < pmax', pmin, p_check, pmax)
                if merge:
                    break
            if merge:
                mod = True
                a = merge[0]
                b = merge[1]
                new_count = p_count[a] + p_count[b]
                #new_p = round(((p_count[a] * a) + (p_count[b] * b)) / new_count)
                new_p = round((a + b) / 2)
                del p_count[a]
                del p_count[b]
                p_count[new_p] = new_count
                p_map[a] = new_p
                p_map[b] = new_p
                for i in p_map:
                    if p_map[i] == a or p_map[i] == b:
                        p_map[i] = new_p # pylint: disable=E4702

        return p_map

    @staticmethod
    def pulses_to_space_encoded_head_key( pulses ):
        results = []
        bits = 0
        data = 0
        byts = []
        start_pulse = 0
        start_space = 0
        last_space = 0

        is_pulse = False
        fail_if_again = False
        for current_ps_time in pulses:
            if fail_if_again:
                return None
            is_pulse = not is_pulse
            if is_pulse:
                print( 'p', current_ps_time )
                if current_ps_time >= 7900 and current_ps_time <= 10100:
                    # NEC protocol start pulse
                    if start_pulse > 0:
                        if not start_space:
                            return None
                        bits2 = (len(byts) * 8) + bits
                        if bits2 == 0 and start_space == 2250:
                            # repeat code
                            pass
                        elif bits2 < 8 or bits2 > 100:
                            return None
                        results.append( (start_pulse, start_space, byts, bits, data, last_space) )
                    start_pulse = 9000
                    bits = data = start_space = last_space = 0
                    byts = []
                elif current_ps_time >= 3400 and current_ps_time <= 5600:
                    # Samsung procol start pulse
                    if start_pulse > 0:
                        if not start_space:
                            return None
                        bits2 = (len(byts) * 8) + bits
                        if bits2 == 0 and start_space == 2250:
                            # repeat code
                            pass
                        elif bits2 < 8 or bits2 > 100:
                            return None
                        results.append( (start_pulse, start_space, byts, bits, data, last_space) )
                    start_pulse = 4500
                    bits = data = start_space = last_space = 0
                    byts = []
                elif current_ps_time > 665 or current_ps_time < 400:
                    # not NEC/Samsung
                    return None
                elif start_space == 0:
                    # not NEC/Samsung
                    return None
            else: # not is_pulse
                print( 's', current_ps_time )
                if start_space == 0:
                    if current_ps_time >= 3400 and current_ps_time <= 5600:
                        # normal start space
                        start_space = 4500
                    elif current_ps_time >= 1150 and current_ps_time <= 3350:
                        # repeat code
                        start_space = 2250
                    else:
                        # not NEC/Samsung
                        return None
                else:
                    if current_ps_time > 3350:
                        # gap between transmissions
                        if start_pulse > 0:
                            if not start_space:
                                return None
                            bits2 = (len(byts) * 8) + bits
                            #if bits2 < 8 or bits2 > 100:
                            #    return None
                            results.append( (start_pulse, start_space, byts, bits, data, last_space) )
                            start_pulse = 9000
                            bits = data = start_space = last_space = 0
                            byts = []
                    elif current_ps_time >= 400 and current_ps_time <= 665:
                        # zero
                        bits += 1
                        if bits == 8:
                            byts.append(data)
                            bits = data = 0
                    elif current_ps_time >= 1400 and current_ps_time <= 1800:
                        # one
                        bits += 1
                        data |= (1 << (8 - bits))
                        if bits == 8:
                            byts.append(data)
                            bits = data = 0
                    else:
                        fail_if_again = True
        if start_pulse > 0:
            if not start_space:
                return None
            bits2 = (len(byts) * 8) + bits
            if bits2 == 0 and start_space == 2250:
                # repeat code
                pass
            elif bits2 < 8 or bits2 > 100:
                return None
            results.append( (start_pulse, start_space, byts, bits, data, last_space) )

        if not results:
            return None

        count = -1
        for r in results:
            count += 1
            if count == 0:
                continue
            # make sure start pulse is the same
            if r[0] != results[0][0]:
                return None

        result_string = ''
        symbols = { results[0][0]: '%', 4500: '^', 2250: '&', results[0][5]: '*' }
        for r in results:
            if r[0] not in symbols or r[1] not in symbols:
                return None
            result_string += symbols[r[0]] + symbols[r[1]]
            result_string += IRRemoteControlDevice._build_key_bitfield( r[3], r[4], r[2] )
            result_string += '@*'

        return result_string

    @staticmethod
    def _build_key_bitfield( bits, bitdata, byts ):
        numbits = bits + (len(byts) * 8)
        result_string = '%02X%02X' % (0, numbits)
        for b in byts:
            result_string += '%02X' % b
        if bits:
            result_string += '%02X' % bitdata
        #print('bitfield:', result_string)
        return result_string

    @staticmethod
    def hex_to_pulses( code_hex ):
        raw_bytes = bytes.fromhex(code_hex)
        return [int.from_bytes(raw_bytes[x:x+2], byteorder="little") for x in range(0, len(raw_bytes), 2)]

    @staticmethod
    def pulses_to_hex( pulses ):
        return "".join([f"{((x >> 8) | (x << 8)) & 0xFFFF:04x}" for x in pulses])

    @staticmethod
    def width_encoded_to_pulses( uint32, start_mark=9000, start_space=4500, pulse_one=563, pulse_zero=563, space_one=1688, space_zero=563, trailing_pulse=563, trailing_space=30000 ):
        pulses = [ start_mark, start_space ]
        one = [ pulse_one, space_one ]
        zero =  [ pulse_zero, space_zero ]
        for i in range(31, -1, -1):
            pulses += one if uint32 & (1 << i) else zero
        pulses.append( trailing_pulse )
        pulses.append( trailing_space )
        return pulses

    @staticmethod
    def pulses_to_width_encoded( pulses, start_mark=None, start_space=None, pulse_threshold=None, space_threshold=None ):
        ret = [ ]
        if len(pulses) < 68:
            log.debug('Length of pulses must be a multiple of 68! (2 start + 64 data + 2 trailing)')
            return ret
        if (pulse_threshold is None) and (space_threshold is None):
            log.debug('"pulse_threshold" and/or "space_threshold" must be supplied!')
            return ret

        if start_mark is not None:
            while( len(pulses) >= 68 and (pulses[0] < (start_mark * 0.75) or pulses[0] > (start_mark * 1.25)) ):
                pulses = pulses[1:]

        while( len(pulses) >= 68 ):
            if start_mark is not None:
                if pulses[0] < (start_mark * 0.75) or pulses[0] > (start_mark * 1.25):
                    log.debug('The start mark is not the correct length')
                    return ret
            if start_space is not None:
                if pulses[1] < (start_space * 0.75) or pulses[1] > (start_space * 1.25):
                    log.debug('The start space is not the correct length')
                    return ret

            pulses = pulses[2:]
            uint32 = 0

            for i in range(31, -1, -1):
                pulse_match = space_match = None
                if pulse_threshold is not None:
                    if pulses[0] >= pulse_threshold:
                        pulse_match = 1
                    else:
                        pulse_match = 0

                if space_threshold is not None:
                    if pulses[1] >= space_threshold:
                        space_match = 1
                    else:
                        space_match = 0

                if (pulse_match is not None) and (space_match is not None):
                    if pulse_match != space_match:
                        log.debug('Both "pulse_threshold" and "space_threshold" are supplied and bit %d conflicts with both!' % i)
                        return ret
                    res = space_match
                elif pulse_match is None:
                    res = space_match
                else:
                    res = pulse_match

                uint32 |= res << i
                pulses = pulses[2:]

            pulses = pulses[2:]

            if ret is None:
                ret = [ uint32 ]
            elif uint32 not in ret:
                ret.append( uint32 )

        return ret

    @staticmethod
    def _mirror_bits( data, bits=8 ):
        shift = bits - 1
        out = 0
        for i in range(bits):
            if data & (1 << i):
                out |= 1 << shift
            shift -= 1
        return out

    @staticmethod
    def nec_to_pulses( address, data=None ):
        # address can be 8-bit or 16-bit
        # if 8, it is repeated after complementing (just like the data)
        if data is None:
            uint32 = address
        else:
            if address < 256:
                address = IRRemoteControlDevice._mirror_bits(address)
                address = (address << 8) | (address ^ 0xFF)
            else:
                address = (IRRemoteControlDevice._mirror_bits( (address >> 8) & 0xFF) << 8) | IRRemoteControlDevice._mirror_bits(address & 0xFF)
            data = IRRemoteControlDevice._mirror_bits(data)
            data = (data << 8) | (data ^ 0xFF)
            uint32 = (address << 16) | data
        return IRRemoteControlDevice.width_encoded_to_pulses( uint32 )

    @staticmethod
    def pulses_to_nec( pulses ):
        ret = [ ]
        res = IRRemoteControlDevice.pulses_to_width_encoded( pulses, start_mark=9000, space_threshold=1125 )
        for code in res:
            addr = IRRemoteControlDevice._mirror_bits((code >> 24) & 0xFF)
            addr_not = IRRemoteControlDevice._mirror_bits((code >> 16) & 0xFF)
            data = IRRemoteControlDevice._mirror_bits((code >> 8) & 0xFF)
            data_not = IRRemoteControlDevice._mirror_bits(code & 0xFF)
            # if the address is 8-bit, it is repeated after complementing (just like the data)
            if addr != (addr_not ^ 0xFF):
                addr = (addr << 8) | addr_not
            d = { 'type': 'nec', 'uint32': code, 'address': None, 'data': None, 'hex': '%08X' % code }
            if data == (data_not ^ 0xFF):
                d['address'] = addr
                d['data'] = data
            ret.append(d)
        return ret

    @staticmethod
    def samsung_to_pulses( address, data=None ):
        if data is None:
            uint32 = address
        else:
            address = IRRemoteControlDevice._mirror_bits(address)
            data = IRRemoteControlDevice._mirror_bits(data)
            uint32 = (address << 24) + (address << 16) + (data << 8) + (data ^ 0xFF)
        return IRRemoteControlDevice.width_encoded_to_pulses( uint32, start_mark=4500 )

    @staticmethod
    def pulses_to_samsung( pulses ):
        ret = [ ]
        res = IRRemoteControlDevice.pulses_to_width_encoded( pulses, start_mark=4500, space_threshold=1125 )
        for code in res:
            addr = (code >> 24) & 0xFF
            addr_not = (code >> 16) & 0xFF
            data = (code >> 8) & 0xFF
            data_not = code & 0xFF
            d = { 'type': 'samsung', 'uint32': code, 'address': None, 'data': None, 'hex': '%08X' % code }
            # samsung repeats the 8-bit address but complements the 8-bit data
            if addr == addr_not and data == (data_not ^ 0xFF):
                d['address'] = IRRemoteControlDevice._mirror_bits(addr)
                d['data'] = IRRemoteControlDevice._mirror_bits(data)
            ret.append(d)
        return ret

    @staticmethod
    def pronto_to_pulses( pronto ):
        ret = [ ]
        pronto = [int(x, 16) for x in pronto.split(' ')]
        ptype = pronto[0]
        timebase = pronto[1]
        pair1_len = pronto[2]
        pair2_len = pronto[3]
        if ptype != 0:
            # only raw (learned) codes are handled
            return ret
        if timebase < 90 or timebase > 139:
            # only 38 kHz is supported?
            return ret
        pronto = pronto[4:]
        timebase *= 0.241246
        for i in range(0, pair1_len*2, 2):
            ret += [round(pronto[i] * timebase), round(pronto[i+1] * timebase)]
        pronto = pronto[pair1_len*2:]
        for i in range(0, pair2_len*2, 2):
            ret += [round(pronto[i] * timebase), round(pronto[i+1] * timebase)]
        return ret

    @staticmethod
    def pulses_to_pronto( pulses ):
        # only 38 kHz is supported?
        freq = 38000.0
        scale = (1 / freq) * 1000000.0
        ret = '%04X %04X %04X %04X' % (0, round(scale/0.241246), 0, len(pulses) >> 1)
        for i in pulses:
            ret += ' %04X' % round(i/scale)
        return ret

    @staticmethod
    def pronto_to_head_key( pronto ):
        ret = [ ]
        pronto = [int(x, 16) for x in pronto.split(' ')]
        ptype = pronto[0]
        timebase = pronto[1]
        pair1_len = pronto[2]
        pair2_len = pronto[3]
        if ptype != 0:
            # only raw (learned) codes are handled
            return None

        # 4,145,152 is 32,768 * 506 / 4
        freq = round(4145152.0 / timebase / 100) / 10

        pronto = pronto[4:]
        timebase *= 0.241246
        for i in range(0, pair1_len*2, 2):
            ret += [round(pronto[i] * timebase), round(pronto[i+1] * timebase)]
        pronto = pronto[pair1_len*2:]
        for i in range(0, pair2_len*2, 2):
            ret += [round(pronto[i] * timebase), round(pronto[i+1] * timebase)]

        return IRRemoteControlDevice.pulses_to_head_key( ret, freq=freq )



# TinyTuya Bulb Device
# -*- coding: utf-8 -*-
"""
 Python module to interface with Tuya WiFi smart devices

 Author: Jason A. Cox
 For more information see https://github.com/jasonacox/tinytuya

 Local Control Classes
    BulbDevice(...)
        See OutletDevice() for constructor arguments

 Functions
    BulbDevice Class methods
        rgb_to_hexvalue(r, g, b, hexformat):
        hsv_to_hexvalue(h, s, v, hexformat):
        hexvalue_to_rgb(hexvalue, hexformat=None):
        hexvalue_to_hsv(hexvalue, hexformat=None):

    BulbDevice
        set_mode(self, mode="white", nowait=False):
        set_scene(self, scene, scene_data=None, nowait=False):
        set_timer(self, num_secs, nowait=False):
        set_musicmode(self, transition, modify_settings=True, nowait=False):
        unset_musicmode( self ):
        set_music_colour( self, red, green, blue, brightness=None, colourtemp=None, transition=None, nowait=False ):
        set_colour(r, g, b, nowait):
        set_hsv(h, s, v, nowait):
        set_white_percentage(brightness=100, colourtemp=0, nowait):
        set_brightness_percentage(brightness=100, nowait):
        set_colourtemp_percentage(colourtemp=100, nowait):
        result = get_value(self, feature, state=None, nowait=False):
        result = get_mode(self, state=None, nowait=False):
        result = get_brightness_percentage(self, state=None, nowait=False):
        result = get_colourtemp_percentage(self, state=None, nowait=False):
        (r, g, b) = colour_rgb():
        (h,s,v) = colour_hsv()
        result = state():
        bool = bulb_has_capability( self, feature, nowait=False ):
        detect_bulb(self, response=None, nowait=False):
        set_bulb_type(self, bulb_type=None, mapping=None):
        set_bulb_capabilities(self, mapping):

    Inherited
        Every device function from core.py
"""

import colorsys

from .core import Device, log
from .core import error_json, ERR_JSON, ERR_FUNCTION # ERR_RANGE, ERR_STATE, ERR_TIMEOUT

# pylint: disable=R0904
class BulbDevice(Device):
    """
    Represents a Tuya based Smart Light/Bulb.
    """

    DPS_MODE_WHITE = "white"
    DPS_MODE_COLOUR = "colour"
    DPS_MODE_SCENE = "scene"
    DPS_MODE_MUSIC = "music"
    DPS_MODE_SCENE_1 = "scene_1"  # nature
    DPS_MODE_SCENE_2 = "scene_2"
    DPS_MODE_SCENE_3 = "scene_3"  # rave
    DPS_MODE_SCENE_4 = "scene_4"  # rainbow

    BULB_FEATURE_MODE = 'mode'
    BULB_FEATURE_BRIGHTNESS = 'brightness'
    BULB_FEATURE_COLOURTEMP = 'colourtemp'
    BULB_FEATURE_COLOUR = 'colour'
    BULB_FEATURE_SCENE = 'scene'
    BULB_FEATURE_SCENE_DATA = 'scene_data'
    BULB_FEATURE_TIMER = 'timer'
    BULB_FEATURE_MUSIC = 'music'

    MUSIC_TRANSITION_JUMP = 0
    MUSIC_TRANSITION_FADE = 1

    DEFAULT_DPSET = {}
    DEFAULT_DPSET['A'] = {
        'switch': 1,
        'mode': 2,
        'brightness': 3,
        'colourtemp': 4,
        'colour': 5,
        'scene': 6,
        'scene_data': None, # Type A sets mode to 'scene_N'
        'timer': 7,
        'music': 8,
        'value_min': 25,
        'value_max': 255,
        'value_hexformat': 'rgb8',
    }
    DEFAULT_DPSET['B'] = {
        'switch': 20,
        'mode': 21,
        'brightness': 22,
        'colourtemp': 23,
        'colour': 24,
        'scene': 25,
        'scene_data': 25, # Type B prefixes scene data with idx
        'timer': 26,
        'music': 28,
        'value_min': 10,
        'value_max': 1000,
        'value_hexformat': 'hsv16',
    }
    DEFAULT_DPSET['C'] = {
        'switch': 1,
        'mode': None,
        'brightness': 2,
        'colourtemp': 3,
        'colour': None,
        'scene': None,
        'scene_data': None,
        'timer': None,
        'music': None,
        'value_min': 25,
        'value_max': 255,
        'value_hexformat': 'rgb8',
    }
    DEFAULT_DPSET['None'] = {
        'switch': 1,
        'mode': None,
        'brightness': None,
        'colourtemp': None,
        'colour': None,
        'scene': None,
        'scene_data': None,
        'timer': None,
        'music': None,
        'value_min': 0,
        'value_max': 255,
        'value_hexformat': 'rgb8',
    }


    # These attributes are obsolete and only kept for backwards compatibility
    DPS_INDEX_SETS = [20, 1] # starts at either DP 20 (Type B) or 1 (all others)
    DPS_INDEX_ON = {"A": "1", "B": "20", "C": "1"}
    DPS_INDEX_MODE = {"A": "2", "B": "21", "C": "1"}
    DPS_INDEX_BRIGHTNESS = {"A": "3", "B": "22", "C": "2"}
    DPS_INDEX_COLOURTEMP = {"A": "4", "B": "23", "C": None}
    DPS_INDEX_COLOUR = {"A": "5", "B": "24", "C": None}
    DPS_INDEX_SCENE = {"A": "2", "B": "25", "C": None}
    DPS_INDEX_TIMER = {"A": None, "B": "26", "C": None}
    DPS_INDEX_MUSIC = {"A": None, "B": "27", "C": None}
    DPS = "dps"

    def __init__(self, *args, **kwargs):
        # Set Default Bulb Types
        self.bulb_configured = False
        self.bulb_type = None
        self.has_brightness = None
        self.has_colourtemp = None
        self.has_colour = None
        self.tried_status = False
        self.dpset = {
            'switch': None,
            'mode': None,
            'brightness': None,
            'colourtemp': None,
            'colour': None,
            'scene': None,
            'scene_data': None,
            'timer': None,
            'music': None,
            'value_min': -1,
            'value_max': -1,
            'value_hexformat': 'hsv16',
        }

        # set the default version to None so we do not immediately connect and call status()
        if 'version' not in kwargs or not kwargs['version']:
            kwargs['version'] = None
        super(BulbDevice, self).__init__(*args, **kwargs)

    def status(self, nowait=False):
        result = super(BulbDevice, self).status(nowait=nowait)
        self.tried_status = True
        if result and (not self.bulb_configured) and ('dps' in result):
            self.detect_bulb(result, nowait=nowait)
        return result

    @staticmethod
    def rgb_to_hexvalue(r, g, b, hexformat):
        """
        Convert an RGB value to the hex representation expected by Tuya Bulb.

        While r, g and b are just hexadecimal values of the corresponding
        Red, Green and Blue values, the h, s and v values (which are values
        between 0 and 1) are scaled:
            hexformat="rgb8": 360 (h) and 255 (s and v)
            hexformat="hsv16": 360 (h) and 1000 (s and v)

        Args:
            r(int): Value for the colour red as int from 0-255.
            g(int): Value for the colour green as int from 0-255.
            b(int): Value for the colour blue as int from 0-255.
            hexformat(str): Selects the return format
                "rgb8": rrggbb0hhhssvv
                "hsv16": hhhhssssvvvv
        """
        err = ''
        if not 0 <= r <= 255.0:
            err += '/red'
        if not 0 <= g <= 255.0:
            err += '/green'
        if not 0 <= b <= 255.0:
            err += '/blue'
        if err:
            raise ValueError('rgb_to_hexvalue: The value for %s needs to be between 0 and 255.' % err[1:])

        hsv = colorsys.rgb_to_hsv(r / 255.0, g / 255.0, b / 255.0)

        if hexformat == 'rgb8':
            # r:0-255,g:0-255,b:0-255|rgb|
            hexvalue = '%02x%02x%02x' % (r, g, b)
            # h:0-360,s:0-255,v:0-255|hsv|
            hexvalue += '%04x%02x%02x' % (int(hsv[0] * 360), int(hsv[1] * 255), int(hsv[2] * 255))
        elif hexformat == 'hsv16':
            # h:0-360,s:0-1000,v:0-1000|hsv|
            hexvalue = '%04x%04x%04x' % (int(hsv[0] * 360), int(hsv[1] * 1000), int(hsv[2] * 1000))
        else:
            raise ValueError('rgb_to_hexvalue: hexformat must be either "rgb8" or "hsv16"')

        return hexvalue

    # Deprecated. Kept for backwards compatibility
    @staticmethod
    def _rgb_to_hexvalue(r, g, b, bulb="A"):
        if bulb == "A":
            hexformat = 'rgb8'
        elif bulb == "B":
            hexformat = 'hsv16'
        else:
            # Unsupported bulb type
            raise ValueError("Unsupported bulb type %r - unable to determine hexvalue." % bulb)

        return BulbDevice.rgb_to_hexvalue(r, g, b, hexformat)

    @staticmethod
    def hsv_to_hexvalue(h, s, v, hexformat):
        """
        Convert an HSV value to the hex representation expected by Tuya Bulb.

        Args:
            h(float): colour Hue as float from 0-1
            s(float): colour Saturation as float from 0-1
            v(float): colour Value as float from 0-1
            hexformat(str): Selects the return format
                "rgb8": rrggbb0hhhssvv
                "hsv16": hhhhssssvvvv
        """
        err = ''
        if not 0 <= h <= 1.0:
            err += '/Hue'
        if not 0 <= s <= 1.0:
            err += '/Saturation'
        if not 0 <= v <= 1.0:
            err += '/Value'

        if err:
            raise ValueError( 'hsv_to_hexvalue: The value for %s needs to be between 0 and 1.' % err[1:])

        if hexformat == 'rgb8':
            (r, g, b) = colorsys.hsv_to_rgb(h, s, v)
            return BulbDevice.rgb_to_hexvalue( r * 255.0, g * 255.0, b * 255.0, hexformat )
        elif hexformat == 'hsv16':
            # h:0-360,s:0-1000,v:0-1000|hsv|
            hexvalue = '%04x%04x%04x' % (int(h * 360), int(s * 1000), int(v * 1000))
            return hexvalue
        else:
            raise ValueError('hsv_to_hexvalue: hexformat must be either "rgb8" or "hsv16"')

    @staticmethod
    def hexvalue_to_rgb(hexvalue, hexformat=None):
        """
        Converts the hexvalue used by Tuya for colour representation into
        an RGB value.

        Args:
            hexvalue(string): The hex representation generated by BulbDevice.rgb_to_hexvalue()
            hexformat(str or None):
                "rgb8": The hex is in rrggbb0hhhssvv format
                "hsv16": The hex is in hhhhssssvvvv format
                None: Try to auto-detect the format
        """
        hexvalue_len = len(hexvalue)
        if not hexformat:
            if hexvalue_len == 6 or hexvalue_len == 14:
                hexformat = 'rgb8'
            elif hexvalue_len == 12:
                hexformat = 'hsv16'
            else:
                # Unsupported bulb type
                raise ValueError("Unable to detect hexvalue format. Value string must have 6, 12 or 14 hex digits.")

        if hexformat == 'rgb8':
            if hexvalue_len < 6:
                raise ValueError("RGB value string must have 6 or 14 hex digits.")
            r = int(hexvalue[0:2], 16)
            g = int(hexvalue[2:4], 16)
            b = int(hexvalue[4:6], 16)
        elif hexformat == 'hsv16':
            # hexvalue is in hsv
            if hexvalue_len < 12:
                raise ValueError("HSV value string must have 12 hex digits.")
            h = float(int(hexvalue[0:4], 16) / 360.0)
            s = float(int(hexvalue[4:8], 16) / 1000.0)
            v = float(int(hexvalue[8:12], 16) / 1000.0)
            rgb = colorsys.hsv_to_rgb(h, s, v)
            r = int(rgb[0] * 255)
            g = int(rgb[1] * 255)
            b = int(rgb[2] * 255)
        else:
            raise ValueError('hexvalue_to_rgb: hexformat must be None, "rgb8" or "hsv16"')

        return (r, g, b)

    # Deprecated. Kept for backwards compatibility
    @staticmethod
    def _hexvalue_to_rgb(hexvalue, bulb="A"):
        if bulb == "A":
            hexformat = 'rgb8'
        elif bulb == "B":
            hexformat = 'hsv16'
        else:
            # Unsupported bulb type, attempt to auto-detect format
            hexformat = None
        return BulbDevice.hexvalue_to_rgb(hexvalue, hexformat)

    @staticmethod
    def hexvalue_to_hsv(hexvalue, hexformat=None):
        """
        Converts the hexvalue used by Tuya for colour representation into
        an HSV value.

        Args:
            hexvalue(string): The hex representation generated by BulbDevice.rgb_to_hexvalue()
            hexformat(str or None):
                "rgb8": The hex is in rrggbb0hhhssvv format
                "hsv16": The hex is in hhhhssssvvvv format
                None: Try to auto-detect the format
        """
        hexvalue_len = len(hexvalue)
        if not hexformat:
            if hexvalue_len == 6 or hexvalue_len == 14:
                hexformat = 'rgb8'
            elif hexvalue_len == 12:
                hexformat = 'hsv16'
            else:
                # Unsupported bulb type
                raise ValueError("Unable to detetect hexvalue format. Value string must have 6, 12 or 14 hex digits.")

        if hexformat == 'rgb8':
            if hexvalue_len < 6:
                raise ValueError("RGB[HSV] value string must have 6 or 14 hex digits.")
            if hexvalue_len < 14:
                # hexvalue is in rgb
                rgb = BulbDevice.hexvalue_to_rgb(hexvalue, 'rgb8')
                h, s, v = colorsys.rgb_to_hsv(rgb[0] / 255.0, rgb[1] / 255.0, rgb[2] / 255.0)
            else:
                # hexvalue is in rgb+hsv
                h = int(hexvalue[7:10], 16) / 360.0
                s = int(hexvalue[10:12], 16) / 255.0
                v = int(hexvalue[12:14], 16) / 255.0
        elif hexformat == 'hsv16':
            # hexvalue is in hsv
            if hexvalue_len < 12:
                raise ValueError("HSV value string must have 12 hex digits.")
            h = int(hexvalue[0:4], 16) / 360.0
            s = int(hexvalue[4:8], 16) / 1000.0
            v = int(hexvalue[8:12], 16) / 1000.0
        else:
            raise ValueError('hexvalue_to_hsv: hexformat must be None, "rgb8" or "hsv16"')

        return (h, s, v)

    # Deprecated. Kept for backwards compatibility
    @staticmethod
    def _hexvalue_to_hsv(hexvalue, bulb="A"):
        if bulb == "A":
            hexformat = 'rgb8'
        elif bulb == "B":
            hexformat = 'hsv16'
        else:
            # Unsupported bulb type, attempt to auto-detect format
            hexformat = None
        return BulbDevice.hexvalue_to_hsv(hexvalue, hexformat)

    def _set_values_check( self, check_values, nowait=False ):
        dps_values = {}

        # check to see which DPs need to be set
        state = self.cached_status(nowait=nowait)
        if state and 'dps' in state and state['dps']:
            # last state is cached, so check to see if 'mode' needs to be set
            for k in check_values:
                dp = self.dpset[k]
                if dp and ((dp not in state['dps']) or (state['dps'][dp] != check_values[k])):
                    dps_values[dp] = check_values[k]
                elif not dp:
                    log.debug('Device does not support capability, skipping: %r:%r', k, check_values[k])

        if dps_values:
            log.debug('Only sending changed DPs: %r', dps_values)
        else:
            # last state not cached or everything already set, so send them all
            for k in check_values:
                if self.dpset[k]:
                    dps_values[self.dpset[k]] = check_values[k]
                else:
                    log.debug('Device does not support capability, skipping: %r:%r', k, check_values[k])
            log.debug('No DPs have changed, sending full update to refresh: %r', dps_values)

        return self.set_multiple_values( dps_values, nowait=nowait )

    def turn_onoff(self, on, switch=0, nowait=False):
        """Turn the device on or off"""
        if not switch:
            if not self.tried_status:
                self.detect_bulb( nowait=nowait )
            # some people may use BulbDevice as the default even for non-bulb
            #   devices, so default to '1' if we can't detect it
            switch = self.dpset['switch'] if self.dpset['switch'] else 1
        return self.set_status(on, switch, nowait=nowait)

    def turn_on(self, switch=0, nowait=False):
        """Turn the device on"""
        return self.turn_onoff( True, switch=switch, nowait=nowait )

    def turn_off(self, switch=0, nowait=False):
        """Turn the device off"""
        return self.turn_onoff( False, switch=switch, nowait=nowait )

    def set_mode(self, mode="white", nowait=False):
        """
        Set bulb mode

        Args:
            mode(string): white,colour,scene,music
            nowait(bool): True to send without waiting for response.
        """
        if not self.bulb_has_capability( 'mode', nowait=nowait ):
            return error_json(ERR_FUNCTION, 'Bulb does not support mode setting.')

        check_values = {
            'mode': mode,
            'switch': True,
        }

        return self._set_values_check( check_values, nowait=nowait )

    def set_scene(self, scene, scene_data=None, nowait=False):
        """
        Set to scene mode

        Args:
            scene(int): Value for the scene as int from 1-4 (Type A bulbs) or 1-N (Type B bulbs).
            nowait(bool): True to send without waiting for response.
        """
        if not self.bulb_has_capability( 'scene', nowait=nowait ):
            return error_json(ERR_FUNCTION, 'set_scene: Bulb does not support scenes.')

        # Type A, scene idx is part of the mode
        if (not self.dpset['scene_data']) or (self.dpset['scene_data'] == self.dpset['mode']):
            if (not 1 <= scene <= 4):
                raise ValueError('set_scene: The value for scene needs to be between 1 and 4.')
            dps_values = {
                self.dpset['mode']: self.DPS_MODE_SCENE + '_' + str(scene)
            }
        else:
            scene = '%02x' % int(scene)
            dps_values = {
                'scene': scene,
                'mode': self.DPS_MODE_SCENE,
            }

            if scene_data:
                if (self.dpset['scene_data'] is True) or (self.dpset['scene_data'] == self.dpset['scene']):
                    dps_values['scene'] += scene_data
                else:
                    dps_values['scene_data'] = scene_data

        return self._set_values_check( dps_values, nowait=nowait )

    def set_timer(self, num_secs, dps_id=0, nowait=False):
        """
        Set the timer

        Args:
            num_secs: data to send to bulb
            dps_id: do not use, kept for compatibility with Device.set_timer()
        """
        if dps_id:
            return self.set_value(dps_id, num_secs, nowait=nowait)

        if not self.bulb_has_capability( 'timer', nowait=nowait ):
            return error_json(ERR_FUNCTION, 'set_timer: Bulb does not support timer.')
        return self.set_value(self.dpset['timer'], num_secs, nowait=nowait)

    def set_music_colour( self, transition, red, green, blue, brightness=None, colourtemp=None, nowait=False ):
        """
        Set a colour while in music mode

        Args:
            red(float): red value, 0.0 - 255.0
            green(float): green value, 0.0 - 255.0
            blue(float): blue value, 0.0 - 255.0
            brightness(float): optional white light brightness
            colourtemp(float): optional white light colourtemp
            transition(int): optional transition. will use transition provided in set_musicmode() if not provided
        """
        if not self.bulb_has_capability( 'music', nowait=nowait ):
            return error_json(ERR_FUNCTION, "set_music_colour: Device does not support music mode.")

        colour = '%x' % transition
        colour += self.rgb_to_hexvalue( red, green, blue, self.dpset['value_hexformat'] )

        if (not brightness) or (brightness < 0):
            brightness = 0
        brightness = int(self.dpset['value_max'] * brightness // 100)

        if (not colourtemp) or (colourtemp < 0):
            colourtemp = 0
        colourtemp = int(self.dpset['value_max'] * colourtemp // 100)

        fmt = '%02x' if self.dpset['value_hexformat'] == 'rgb8' else '%04x'
        colour += fmt % brightness
        colour += fmt % colourtemp

        return self.set_value(self.dpset['music'], colour, nowait=nowait)

    def set_colour(self, r, g, b, nowait=False):
        """
        Set colour of an rgb bulb.

        Args:
            r(float): Value for the colour Red from 0.0-255.0.
            g(float): Value for the colour Green from 0.0-255.0.
            b(float): Value for the colour Blue from 0.0-255.0.
            nowait(bool): True to send without waiting for response.
        """
        if not self.bulb_has_capability( 'colour', nowait=nowait ):
            return error_json(ERR_FUNCTION, "set_colour: Device does not support color.")

        check_values = {
            'colour': self.rgb_to_hexvalue(r, g, b, self.dpset['value_hexformat']),
            'mode': self.DPS_MODE_COLOUR,
            'switch': True
        }

        return self._set_values_check( check_values, nowait=nowait )

    def set_hsv(self, h, s, v, nowait=False):
        """
        Set colour of an rgb bulb using h, s, v.

        Args:
            h(float): colour Hue as float from 0-1
            s(float): colour Saturation as float from 0-1
            v(float): colour Value as float from 0-1
            nowait(bool): True to send without waiting for response.
        """
        if not self.bulb_has_capability( 'colour', nowait=nowait ):
            return error_json(ERR_FUNCTION, "set_colour: Device does not support color.")

        check_values = {
            'colour': self.hsv_to_hexvalue( h, s, v, self.dpset['value_hexformat'] ),
            'mode': self.DPS_MODE_COLOUR,
            'switch': True
        }
        return self._set_values_check( check_values, nowait=nowait )

    def set_white_percentage(self, brightness=100, colourtemp=0, nowait=False):
        """
        Set white coloured theme of an rgb bulb.

        Args:
            brightness(int): Value for the brightness in percent (0-100)
            colourtemp(int): Value for the colour temperature in percent (0-100)
            nowait(bool): True to send without waiting for response.

        Note: unlike set_colourtemp(), the colour temp will be silently ignored if the bulb does not support it
        """
        err = ''
        if not 0 <= brightness <= 100:
            err += '/Brightness'
        if not 0 <= colourtemp <= 100:
            err += '/Colourtemp'
        if err:
            raise ValueError( 'set_white_percentage: %s percentage needs to be between 0 and 100.' % err[1:])

        b = int(self.dpset['value_max'] * brightness // 100)
        c = int(self.dpset['value_max'] * colourtemp // 100)

        return self.set_white( b, c, nowait=nowait )

    # Deprecated.  Please use set_white_percentage() instead.
    def set_white(self, brightness=-1, colourtemp=-1, nowait=False):
        """
        DEPRECATED Set white coloured theme of an rgb bulb.

        Args:
            brightness(int): Value for the brightness (A:25-255 or B:10-1000)
            colourtemp(int): Value for the colour temperature (A:0-255, B:0-1000).
            nowait(bool): True to send without waiting for response.

            Default: Max Brightness and Min Colourtemp

        Note: unlike set_colourtemp(), the colour temp will be silently ignored if the bulb does not support it
        """
        if not self.bulb_has_capability( 'brightness', nowait=nowait ):
            return error_json(ERR_FUNCTION, 'set_white: Device does not support brightness.')

        # Brightness (default: Max)
        brightness = int(brightness)
        if brightness < 0:
            brightness = self.dpset['value_max']
        elif brightness > self.dpset['value_max']:
            raise ValueError('set_white: The brightness needs to be between %d and %d.' % (self.dpset['value_min'], self.dpset['value_max']))

        # Colourtemp (default: Min)
        # It will be silently ignored if the bulb does not support it
        if colourtemp is not None:
            colourtemp = int(colourtemp)
            if colourtemp < 0:
                colourtemp = 0
            if colourtemp > self.dpset['value_max']:
                raise ValueError('set_white: The colour temperature needs to be between 0 and %d.' % self.dpset['value_max'])

        # do this the hard way as brightness=0 means we should turn off, but if colourtemp is set then
        #   turn_on() should turn it on at that colourtemp
        check_values = {}
        if brightness >= self.dpset['value_min']:
            check_values['brightness'] = brightness
        if colourtemp is not None:
            check_values['colourtemp'] = colourtemp
        if check_values:
            # we're setting brightness and/or colourtemp
            check_values['mode'] = self.DPS_MODE_WHITE
        check_values['switch'] = bool(brightness >= self.dpset['value_min'])

        # _set_values_check() will skip colourtemp if the bulb does not have it
        return self._set_values_check( check_values, nowait=nowait )

    def set_brightness_percentage(self, brightness=100, nowait=False):
        """
        Set the brightness value of an rgb bulb.

        Args:
            brightness(int): Value for the brightness in percent (0-100)
            nowait(bool): True to send without waiting for response.
        """
        if not 0 <= brightness <= 100:
            raise ValueError('set_brightness_percentage: The brightness needs to be between 0 and 100.')
        b = int(self.dpset['value_max'] * brightness // 100)
        return self.set_brightness(b, nowait=nowait)

    # Deprecated.  Please use set_brightness_percentage() instead.
    def set_brightness(self, brightness, nowait=False):
        """
        DEPRECATED Set the brightness value of an rgb bulb.

        Args:
            brightness(int): Value for the brightness (25-255).
            nowait(bool): True to send without waiting for response.
        """
        if not self.bulb_has_capability( 'brightness', nowait=nowait ):
            return error_json(ERR_FUNCTION, 'set_brightness: Device does not support brightness.')

        # Brightness (default Max)
        if brightness < 0:
            brightness = self.dpset['value_max']
        elif brightness < self.dpset['value_min']:
            return self.turn_off(0, nowait=nowait)
        elif brightness > self.dpset['value_max']:
            raise ValueError('set_brightness: The brightness needs to be between %d and %d.' % (self.dpset['value_min'], self.dpset['value_max']))

        # Determine which mode bulb is in and adjust accordingly
        state = self.state(nowait=nowait)

        if ('Error' in state) or ('mode' not in state):
            return state

        if state['mode'] != self.DPS_MODE_COLOUR:
            # use white mode, changing to it if needed
            return self.set_white(brightness=brightness, colourtemp=None, nowait=nowait)
        else:
            # for colour mode use hsv to increase brightness
            value = brightness / float(self.dpset['value_max'])
            (h, s, v) = self.colour_hsv(state=state, nowait=nowait)
            return self.set_hsv(h, s, value, nowait=nowait)

    def set_colourtemp_percentage(self, colourtemp=100, nowait=False):
        """
        Set the colour temperature of an rgb bulb.

        Args:
            colourtemp(int): Value for the colour temperature in percentage (0-100).
            nowait(bool): True to send without waiting for response.
        """
        if not 0 <= colourtemp <= 100:
            raise ValueError( 'set_colourtemp_percentage: Colourtemp percentage needs to be between 0 and 100.')
        c = int(self.dpset['value_max'] * colourtemp // 100)
        return self.set_colourtemp( c, nowait=nowait )

    # Deprecated.  Please use set_white_percentage() instead.
    def set_colourtemp(self, colourtemp, nowait=False):
        """
        DEPRECATED Set the colour temperature of an rgb bulb.

        Args:
            colourtemp(int): Value for the colour temperature (0-255).
            nowait(bool): True to send without waiting for response.
        """
        if not self.bulb_has_capability( self.BULB_FEATURE_COLOURTEMP, nowait=nowait ):
            return error_json(ERR_FUNCTION, 'set_colourtemp: Device does not support colourtemp.')

        if not 0 <= colourtemp <= self.dpset['value_max']:
            raise ValueError('set_colourtemp: The colour temperature needs to be between 0 and %d.' % self.dpset['value_max'])

        check_values = {
            'colourtemp': colourtemp,
            'mode': self.DPS_MODE_WHITE,
            'switch': True,
        }

        return self._set_values_check( check_values, nowait=nowait )

    def get_value(self, feature, state=None, nowait=False):
        if not state:
            state = self.state(nowait=nowait)
        if 'Error' in state:
            raise RuntimeError('Error getting device current state.')
        if feature not in state:
            raise ValueError("Unknown parameter %r." % feature)
        return state[feature]

    def get_mode(self, state=None, nowait=False):
        """Return current working mode"""
        return self.get_value('mode', state=state, nowait=nowait)

    def white_percentage(self, state=None, nowait=False):
        if not state:
            state = self.state(nowait=nowait)
        return (self.brightness_percentage(state=state, nowait=nowait), self.colourtemp_percentage(state=state, nowait=nowait))

    #def white(self, state=None, nowait=False):
    #    pass

    def get_brightness_percentage(self, state=None, nowait=False):
        if (not self.dpset['value_max']) or (self.dpset['value_max'] < 1):
            raise RuntimeError("Bulb capabilitiy 'value_max' not set, unable to calculate percentage.")
        return self.brightness(state=state, nowait=nowait) / self.dpset['value_max'] * 100.0

    def brightness(self, state=None, nowait=False):
        """Return brightness value"""
        return self.get_value('brightness', state=state, nowait=nowait)

    def get_colourtemp_percentage(self, state=None, nowait=False):
        if (not self.dpset['value_max']) or (self.dpset['value_max'] < 1):
            raise RuntimeError("Bulb capabilitiy 'value_max' not set, unable to calculate percentage.")
        return self.colourtemp(state=state, nowait=nowait) / self.dpset['value_max'] * 100.0

    def colourtemp(self, state=None, nowait=False):
        """Return colour temperature"""
        return self.get_value('colourtemp', state=state, nowait=nowait)

    def colour_rgb(self, state=None, nowait=False):
        """Return colour as RGB value"""
        hexvalue = self.get_value('colour', state=state, nowait=nowait)
        if isinstance( hexvalue, dict ):
            return hexvalue # Error
        return BulbDevice.hexvalue_to_rgb(hexvalue, self.dpset['value_hexformat'])

    def colour_hsv(self, state=None, nowait=False):
        """Return colour as HSV value"""
        hexvalue = self.get_value('colour', state=state, nowait=nowait)
        if isinstance( hexvalue, dict ):
            return hexvalue # Error
        return BulbDevice.hexvalue_to_hsv(hexvalue, self.dpset['value_hexformat'])

    def state(self, nowait=False):
        """Return state of Bulb"""
        if not self.bulb_configured:
            self.detect_bulb(nowait=nowait)
            if not self.bulb_configured:
                raise RuntimeError('Bulb not configured, cannot get device current state.')

        status = self.cached_status(nowait=nowait)
        state = {}
        if not status:
            return error_json(ERR_JSON, "state: empty response")

        if "Error" in status:
            return error_json(ERR_JSON, status["Error"])

        if 'dps' not in status:
            return error_json(ERR_JSON, "state: no data points")

        for key in self.dpset:
            dp = self.dpset[key]
            if '_' in key:
                # skip scene_data, value_min, value_max, etc
                state[key] = None
            elif dp in status['dps']:
                state[key] = status['dps'][dp]
            else:
                state[key] = None

        if 'switch' in state:
            state['is_on'] = state['switch']

        #print( 'state:', state )
        return state

    def bulb_has_capability( self, feature, nowait=False ):
        if not self.bulb_configured:
            self.detect_bulb( nowait=nowait )
            if not self.bulb_configured:
                raise RuntimeError('Bulb not configured, cannot get device capabilities.')
        return bool( self.dpset[feature] )

    def detect_bulb(self, response=None, nowait=False):
        """
        Attempt to determine BulbDevice Type A, B or C based on:
            Type A has keys 1-9
            Type B has keys 20-28
            Type C is basic (non-CCT) and only has 1-2 (i.e Feit type bulbs from Costco)

        Example status data:
          Sylvania BR30 [v3.3, RGB+CCT]:
            {'20': True, '21': 'colour', '22': 750, '23': 278, '24': '00f003e803e8', '25': '000e0d0000000000000000c803e8', '26': 0}

          Geeni BW229 Smart Filament Bulb [v3.3, CCT only]:
            {'1': True, '2': 25, '3': 0}
            1: switch, 2: brightness, 3: colour temperature

          No-name RGB+CCT (LED BULB W5K) [v3.5, RGB+CCT]:
            {'20': True, '21': 'white', '22': 10, '23': 0, '24': '000003e803e8', '25': '000e0d0000000000000000c80000', '26': 0, '34': False}

          Feit soft white Filament [v3.5, 2700K only]:
            {'20': True, '21': 'white', '22': 60, '25': '000e0d0000000000000000c803e8', '26': 0, '34': False, '41': True}
            (No CCT (23) or colour (24), but does support scenes (25) and music mode (28))

          Feit dimmer switch [v3.3, not a bulb]:
            {'1': True, '2': 10, '3': 10, '4': 'incandescent'}
            Note: after a power cycle, only DP 2 is returned!  The rest are not returned until after they are set
            1: switch, 2: brightness, 3: minimum dim %, 4: installed bulb type (LED/incandescent)
        """
        if not response:
            response = self.cached_status(historic=True, nowait=nowait)
            if (not response) or ('dps' not in response):
                if nowait:
                    log.debug('No cached status, but nowait set! detect_bulb() exiting without detecting bulb!')
                else:
                    response = self.status()
                # return here as self.status() will call us again
                return
        if response and 'dps' in response and isinstance(response['dps'], dict):
            # Try to determine type of BulbDevice Type based on DPS indexes
            # 1+2 or 20+21 are required per https://developer.tuya.com/en/docs/iot/product-function-definition?id=K9tp155s4th6b
            #   The rest are optional
            if '20' in response['dps'] and '1' in response['dps']:
                # both 1 and 20 in response, this probably isn't a bulb
                self.bulb_configured = True
                self.bulb_type = 'None'
            elif '20' in response['dps'] and '21' in response['dps']:
                self.bulb_configured = True
                self.bulb_type = 'B'
            elif '1' in response['dps'] and '2' in response['dps']:
                self.bulb_configured = True

                # if DP 2 is a string, it is the mode (Type A).  If it is an int, it is the brightness (Type C)
                self.bulb_type = 'A' if isinstance(response['dps']['2'], str) else 'C'

            if self.bulb_type and self.bulb_type in self.DEFAULT_DPSET:
                # The 'music' DP is not returned in status(), so use the default value
                self.dpset['music'] = self.DEFAULT_DPSET[self.bulb_type]['music']

                self.dpset['value_min'] = self.DEFAULT_DPSET[self.bulb_type]['value_min']
                self.dpset['value_max'] = self.DEFAULT_DPSET[self.bulb_type]['value_max']
                self.dpset['value_hexformat'] = self.DEFAULT_DPSET[self.bulb_type]['value_hexformat']

                for k in self.DEFAULT_DPSET[self.bulb_type]:
                    if k[:6] == 'value_':
                        continue
                    if not self.DEFAULT_DPSET[self.bulb_type][k]:
                        continue
                    dp = str( self.DEFAULT_DPSET[self.bulb_type][k] )
                    if dp in response['dps']:
                        self.dpset[k] = dp

            # set has_* attributes for backwards compatibility
            for k in ('brightness', 'colourtemp', 'colour'):
                setattr( self, 'has_'+k, bool(self.dpset[k]) )

            log.debug("Bulb type set to %r. has brightness: %r, has colourtemp: %r, has colour: %r",
                      self.bulb_type, self.dpset['brightness'], self.dpset['colourtemp'], self.dpset['colour']
                      )
        elif not self.bulb_configured:
            # response has no dps
            log.debug("No DPs in response, cannot detect bulb type!")

    def set_bulb_type(self, bulb_type=None, mapping=None):
        self.bulb_type = bulb_type
        self.set_bulb_capabilities(mapping)

    def set_bulb_capabilities(self, mapping):
        if self.bulb_type in self.DEFAULT_DPSET:
            default_dpset = self.DEFAULT_DPSET[self.bulb_type]
        else:
            default_dpset = {}

        if not isinstance( mapping, dict ):
            mapping = {}

        for k in self.dpset:
            if k in mapping:
                self.dpset[k] = mapping[k]
            elif self.dpset[k] is None:
                dp = default_dpset.get(k, None)
                self.dpset[k] = str(dp) if (dp and k[:6] != 'value_') else dp

        if self.dpset['switch'] and self.dpset['brightness']:
            self.bulb_configured = True



# TinyTuya Module
# -*- coding: utf-8 -*-

# Globals Network Settings
MAXCOUNT = 15       # How many tries before stopping
SCANTIME = 18       # How many seconds to wait before stopping device discovery
UDPPORT = 6666      # Tuya 3.1 UDP Port
UDPPORTS = 6667     # Tuya 3.3 encrypted UDP Port
UDPPORTAPP = 7000   # Tuya app encrypted UDP Port
TCPPORT = 6668      # Tuya TCP Local Port
TIMEOUT = 3.0       # Seconds to wait for a broadcast
TCPTIMEOUT = 0.4    # Seconds to wait for socket open for scanning
DEFAULT_NETWORK = '192.168.0.0/24'

# Configuration Files
CONFIGFILE = 'tinytuya.json'
DEVICEFILE = 'devices.json'
RAWFILE = 'tuya-raw.json'
SNAPSHOTFILE = 'snapshot.json'

DEVICEFILE_SAVE_VALUES = ('category', 'product_name', 'product_id', 'biz_type', 'model', 'sub', 'icon', 'version', 'last_ip', 'uuid', 'node_id', 'sn', 'mapping')



# TinyTuya Module
# -*- coding: utf-8 -*-

from __future__ import print_function  # python 2.7 support
import base64
import logging
import time

for clib in ('pyca/cryptography', 'PyCryptodomex', 'PyCrypto', 'pyaes'):
    Crypto = Crypto_modes = AES = CRYPTOLIB = None
    try:
        if clib == 'pyca/cryptography': # https://cryptography.io/en/latest/
            from cryptography import __version__ as Crypto_version
            if (Crypto_version[:2] in ('0.', '1.', '2.')) or (Crypto_version == '3.0'):
                # cryptography <= 3.0 requires a backend= parameter
                continue
            from cryptography.hazmat.primitives.ciphers import Cipher as Crypto
            from cryptography.hazmat.primitives.ciphers import modes as Crypto_modes
            from cryptography.hazmat.primitives.ciphers.algorithms import AES
        elif clib == 'PyCryptodomex': # https://pycryptodome.readthedocs.io/en/latest/
            # PyCryptodome is installed as "Cryptodome" when installed by
            #  `apt install python3-pycryptodome` or `pip install pycryptodomex`
            import Cryptodome as Crypto
            from Cryptodome.Cipher import AES
        elif clib == 'PyCrypto': # https://www.pycrypto.org/
            import Crypto
            from Crypto.Cipher import AES
            # v1/v2 is PyCrypto, v3 is PyCryptodome
            clib = 'PyCrypto' if Crypto.version_info[0] < 3 else 'PyCryptodome'
        elif clib == 'pyaes':
            import pyaes  # https://github.com/ricmoo/pyaes
        else:
            continue
        CRYPTOLIB = clib
        break
    except ImportError:
        continue
if CRYPTOLIB is None:
    raise ModuleNotFoundError('No crypto library found, please "pip install" cryptography, pycryptodome, or pyaes')

log = logging.getLogger(__name__)


# Cryptography Helpers
class _AESCipher_Base(object):
    def __init__(self, key):
        self.key = key

    @classmethod
    def get_encryption_iv( cls, iv ):
        if not cls.CRYPTOLIB_HAS_GCM:
            raise NotImplementedError( 'Crypto library does not support GCM' )
        if iv is True:
            if log.isEnabledFor( logging.DEBUG ):
                iv = b'0123456789ab'
            else:
                iv = str(time.time() * 10)[:12].encode('utf8')
        return iv

    @classmethod
    def get_decryption_iv( cls, iv, data ):
        if not cls.CRYPTOLIB_HAS_GCM:
            raise NotImplementedError( 'Crypto library does not support GCM' )
        if iv is True:
            iv = data[:12]
            data = data[12:]
        return iv, data

    @staticmethod
    def _pad(s, bs):
        padnum = bs - len(s) % bs
        return s + padnum * chr(padnum).encode()

    @staticmethod
    def _unpad(s, verify_padding=False):
        padlen = ord(s[-1:])
        if padlen < 1 or padlen > 16:
            raise ValueError("invalid padding length byte")
        if verify_padding and s[-padlen:] != (padlen * chr(padlen).encode()):
            raise ValueError("invalid padding data")
        return s[:-padlen]

class _AESCipher_pyca(_AESCipher_Base):
    def encrypt(self, raw, use_base64=True, pad=True, iv=False, header=None): # pylint: disable=W0621
        if iv: # initialization vector or nonce (number used once)
            iv = self.get_encryption_iv( iv )
            encryptor = Crypto( AES(self.key), Crypto_modes.GCM(iv) ).encryptor()
            if header:
                encryptor.authenticate_additional_data(header)
            crypted_text = encryptor.update(raw) + encryptor.finalize()
            crypted_text = iv + crypted_text + encryptor.tag
        else:
            if pad: raw = self._pad(raw, 16)
            encryptor = Crypto( AES(self.key), Crypto_modes.ECB() ).encryptor()
            crypted_text = encryptor.update(raw) + encryptor.finalize()

        return base64.b64encode(crypted_text) if use_base64 else crypted_text

    def decrypt(self, enc, use_base64=True, decode_text=True, verify_padding=False, iv=False, header=None, tag=None):
        if not iv:
            if use_base64:
                enc = base64.b64decode(enc)
            if len(enc) % 16 != 0:
                raise ValueError("invalid length: %d" % len(enc))
        if iv:
            iv, enc = self.get_decryption_iv( iv, enc )
            if tag is None:
                decryptor = Crypto( AES(self.key), Crypto_modes.CTR(iv + b'\x00\x00\x00\x02') ).decryptor()
            else:
                decryptor = Crypto( AES(self.key), Crypto_modes.GCM(iv, tag) ).decryptor()
            if header and (tag is not None):
                decryptor.authenticate_additional_data( header )
            raw = decryptor.update( enc ) + decryptor.finalize()
        else:
            decryptor = Crypto( AES(self.key), Crypto_modes.ECB() ).decryptor()
            raw = decryptor.update( enc ) + decryptor.finalize()
            raw = self._unpad(raw, verify_padding)
        return raw.decode("utf-8") if decode_text else raw

class _AESCipher_PyCrypto(_AESCipher_Base):
    def encrypt(self, raw, use_base64=True, pad=True, iv=False, header=None): # pylint: disable=W0621
        if iv: # initialization vector or nonce (number used once)
            iv = self.get_encryption_iv( iv )
            cipher = AES.new(self.key, mode=AES.MODE_GCM, nonce=iv)
            if header:
                cipher.update(header)
            crypted_text, tag = cipher.encrypt_and_digest(raw)
            crypted_text = cipher.nonce + crypted_text + tag
        else:
            if pad: raw = self._pad(raw, 16)
            cipher = AES.new(self.key, mode=AES.MODE_ECB)
            crypted_text = cipher.encrypt(raw)

        return base64.b64encode(crypted_text) if use_base64 else crypted_text

    def decrypt(self, enc, use_base64=True, decode_text=True, verify_padding=False, iv=False, header=None, tag=None):
        if not iv:
            if use_base64:
                enc = base64.b64decode(enc)
            if len(enc) % 16 != 0:
                raise ValueError("invalid length: %d" % len(enc))
        if iv:
            iv, enc = self.get_decryption_iv( iv, enc )
            cipher = AES.new(self.key, AES.MODE_GCM, nonce=iv)
            if header:
                cipher.update(header)
            if tag:
                raw = cipher.decrypt_and_verify(enc, tag)
            else:
                raw = cipher.decrypt(enc)
        else:
            cipher = AES.new(self.key, AES.MODE_ECB)
            raw = cipher.decrypt(enc)
            raw = self._unpad(raw, verify_padding)
        return raw.decode("utf-8") if decode_text else raw

class _AESCipher_pyaes(_AESCipher_Base):
    def encrypt(self, raw, use_base64=True, pad=True, iv=False, header=None): # pylint: disable=W0621
        if iv:
            # GCM required for 3.5 devices
            raise NotImplementedError( 'pyaes does not support GCM, please install PyCryptodome' )

        # pylint: disable-next=used-before-assignment
        cipher = pyaes.blockfeeder.Encrypter(
            pyaes.AESModeOfOperationECB(self.key),
            pyaes.PADDING_DEFAULT if pad else pyaes.PADDING_NONE
        )  # no IV, auto pads to 16
        crypted_text = cipher.feed(raw)
        crypted_text += cipher.feed()  # flush final block
        return base64.b64encode(crypted_text) if use_base64 else crypted_text

    def decrypt(self, enc, use_base64=True, decode_text=True, verify_padding=False, iv=False, header=None, tag=None):
        if iv:
            # GCM required for 3.5 devices
            raise NotImplementedError( 'pyaes does not support GCM, please install PyCryptodome' )

        if use_base64:
            enc = base64.b64decode(enc)

        if len(enc) % 16 != 0:
            raise ValueError("invalid length: %d" % len(enc))

        cipher = pyaes.blockfeeder.Decrypter(
            pyaes.AESModeOfOperationECB(self.key),
            pyaes.PADDING_NONE if verify_padding else pyaes.PADDING_DEFAULT
        )  # no IV, auto pads to 16

        raw = cipher.feed(enc)
        raw += cipher.feed()  # flush final block

        if verify_padding: raw = self._unpad(raw, verify_padding)
        return raw.decode("utf-8") if decode_text else raw

if CRYPTOLIB[:8] == 'PyCrypto': # PyCrypto, PyCryptodome, and PyCryptodomex
    class AESCipher(_AESCipher_PyCrypto):
        CRYPTOLIB = CRYPTOLIB
        CRYPTOLIB_VER = '.'.join( [str(x) for x in Crypto.version_info] )
        CRYPTOLIB_HAS_GCM = getattr( AES, 'MODE_GCM', False ) # only PyCryptodome supports GCM, PyCrypto does not
elif CRYPTOLIB == 'pyaes':
    class AESCipher(_AESCipher_pyaes):
        CRYPTOLIB = CRYPTOLIB
        CRYPTOLIB_VER = '.'.join( [str(x) for x in pyaes.VERSION] )
        CRYPTOLIB_HAS_GCM = False
elif CRYPTOLIB == 'pyca/cryptography':
    class AESCipher(_AESCipher_pyca):
        CRYPTOLIB = CRYPTOLIB
        CRYPTOLIB_VER = Crypto_version
        CRYPTOLIB_HAS_GCM = getattr( Crypto_modes, 'GCM', False )
